# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Assignment service."""

import copy
import itertools

from oslo_log import log

from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
from keystone.common.resource_options import options as ro_opt
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone import notifications

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs

# This is a general cache region for assignment administration (CRUD
# operations).
MEMOIZE = cache.get_memoization_decorator(group='role')

# This builds a discrete cache region dedicated to role assignments computed
# for a given user + project/domain pair. Any write operation to add or remove
# any role assignment should invalidate this entire cache region.
COMPUTED_ASSIGNMENTS_REGION = cache.create_region(name='computed assignments')
MEMOIZE_COMPUTED_ASSIGNMENTS = cache.get_memoization_decorator(
    group='role', region=COMPUTED_ASSIGNMENTS_REGION
)


@notifications.listener
class Manager(manager.Manager):
    """Default pivot point for the Assignment backend.

    See :class:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.assignment'
    _provides_api = 'assignment_api'

    _SYSTEM_SCOPE_TOKEN = 'system'  # nosec
    _USER_SYSTEM = 'UserSystem'
    _GROUP_SYSTEM = 'GroupSystem'
    _PROJECT = 'project'
    _ROLE_REMOVED_FROM_USER = 'role_removed_from_user'
    _INVALIDATION_USER_PROJECT_TOKENS = 'invalidate_user_project_tokens'

    def __init__(self):
        assignment_driver = CONF.assignment.driver
        super().__init__(assignment_driver)

        self.event_callbacks = {
            notifications.ACTIONS.deleted: {
                'domain': [self._delete_domain_assignments]
            }
        }

    def _delete_domain_assignments(
        self, service, resource_type, operations, payload
    ):
        domain_id = payload['resource_info']
        self.driver.delete_domain_assignments(domain_id)

    def _get_group_ids_for_user_id(self, user_id):
        # TODO(morganfainberg): Implement a way to get only group_ids
        # instead of the more expensive to_dict() call for each record.
        return [
            x['id']
            for x in PROVIDERS.identity_api.list_groups_for_user(user_id)
        ]

    def list_user_ids_for_project(self, project_id):
        PROVIDERS.resource_api.get_project(project_id)
        assignment_list = self.list_role_assignments(
            project_id=project_id, effective=True
        )
        # Use set() to process the list to remove any duplicates
        return list({x['user_id'] for x in assignment_list})

    def _send_app_cred_notification_for_role_removal(self, role_id):
        """Delete all application credential for a specific role.

        :param role_id: role identifier
        :type role_id: string
        """
        assignments = self.list_role_assignments(role_id=role_id)
        for assignment in assignments:
            if 'user_id' in assignment and 'project_id' in assignment:
                payload = {
                    'user_id': assignment['user_id'],
                    'project_id': assignment['project_id'],
                }
                notifications.Audit.internal(
                    notifications.REMOVE_APP_CREDS_FOR_USER, payload
                )

    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def get_roles_for_user_and_project(self, user_id, project_id):
        """Get the roles associated with a user within given project.

        This includes roles directly assigned to the user on the
        project, as well as those by virtue of group membership or
        inheritance.

        :returns: a list of role ids.
        :raises keystone.exception.ProjectNotFound: If the project doesn't
            exist.

        """
        PROVIDERS.resource_api.get_project(project_id)
        assignment_list = self.list_role_assignments(
            user_id=user_id, project_id=project_id, effective=True
        )
        # Use set() to process the list to remove any duplicates
        return list({x['role_id'] for x in assignment_list})

    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def get_roles_for_trustor_and_project(self, trustor_id, project_id):
        """Get the roles associated with a trustor within given project.

        This includes roles directly assigned to the trustor on the
        project, as well as those by virtue of group membership or
        inheritance, but it doesn't include the domain roles.

        :returns: a list of role ids.
        :raises keystone.exception.ProjectNotFound: If the project doesn't
            exist.

        """
        PROVIDERS.resource_api.get_project(project_id)
        assignment_list = self.list_role_assignments(
            user_id=trustor_id,
            project_id=project_id,
            effective=True,
            strip_domain_roles=False,
        )
        # Use set() to process the list to remove any duplicates
        return list({x['role_id'] for x in assignment_list})

    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def get_roles_for_user_and_domain(self, user_id, domain_id):
        """Get the roles associated with a user within given domain.

        :returns: a list of role ids.
        :raises keystone.exception.DomainNotFound: If the domain doesn't exist.

        """
        PROVIDERS.resource_api.get_domain(domain_id)
        assignment_list = self.list_role_assignments(
            user_id=user_id, domain_id=domain_id, effective=True
        )
        # Use set() to process the list to remove any duplicates
        return list({x['role_id'] for x in assignment_list})

    def get_roles_for_groups(self, group_ids, project_id=None, domain_id=None):
        """Get a list of roles for this group on domain and/or project."""
        # if no group ids were passed, there are no roles. Without this check,
        # all assignments for the project or domain will be fetched,
        # which is not what we want.
        if not group_ids:
            return []
        if project_id is not None:
            PROVIDERS.resource_api.get_project(project_id)
            assignment_list = self.list_role_assignments(
                source_from_group_ids=group_ids,
                project_id=project_id,
                effective=True,
            )
        elif domain_id is not None:
            assignment_list = self.list_role_assignments(
                source_from_group_ids=group_ids,
                domain_id=domain_id,
                effective=True,
            )
        else:
            raise AttributeError(_("Must specify either domain or project"))

        role_ids = list({x['role_id'] for x in assignment_list})
        return PROVIDERS.role_api.list_roles_from_ids(role_ids)

    @notifications.role_assignment('created')
    def _add_role_to_user_and_project_adapter(
        self,
        role_id,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
        context=None,
    ):
        # The parameters for this method must match the parameters for
        # create_grant so that the notifications.role_assignment decorator
        # will work.

        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.role_api.get_role(role_id)
        self.driver.add_role_to_user_and_project(user_id, project_id, role_id)

    def add_role_to_user_and_project(self, user_id, project_id, role_id):
        self._add_role_to_user_and_project_adapter(
            role_id, user_id=user_id, project_id=project_id
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    # TODO(henry-nash): We might want to consider list limiting this at some
    # point in the future.
    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def list_projects_for_user(self, user_id):
        # FIXME(lbragstad): Without the use of caching, listing effective role
        # assignments is slow, especially with large data set (lots of users
        # with multiple role assignments). This should serve as a marker in
        # case we have the opportunity to come back and optimize this code so
        # that it can be performant without having a hard dependency on
        # caching. Please see https://bugs.launchpad.net/keystone/+bug/1700852
        # for more details.
        assignment_list = self.list_role_assignments(
            user_id=user_id, effective=True
        )
        # Use set() to process the list to remove any duplicates
        project_ids = list(
            {x['project_id'] for x in assignment_list if x.get('project_id')}
        )
        return PROVIDERS.resource_api.list_projects_from_ids(project_ids)

    # TODO(henry-nash): We might want to consider list limiting this at some
    # point in the future.
    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def list_domains_for_user(self, user_id):
        assignment_list = self.list_role_assignments(
            user_id=user_id, effective=True
        )
        # Use set() to process the list to remove any duplicates
        domain_ids = list(
            {x['domain_id'] for x in assignment_list if x.get('domain_id')}
        )
        return PROVIDERS.resource_api.list_domains_from_ids(domain_ids)

    def list_domains_for_groups(self, group_ids):
        assignment_list = self.list_role_assignments(
            source_from_group_ids=group_ids, effective=True
        )
        domain_ids = list(
            {x['domain_id'] for x in assignment_list if x.get('domain_id')}
        )
        return PROVIDERS.resource_api.list_domains_from_ids(domain_ids)

    def list_projects_for_groups(self, group_ids):
        assignment_list = self.list_role_assignments(
            source_from_group_ids=group_ids, effective=True
        )
        project_ids = list(
            {x['project_id'] for x in assignment_list if x.get('project_id')}
        )
        return PROVIDERS.resource_api.list_projects_from_ids(project_ids)

    @notifications.role_assignment('deleted')
    def _remove_role_from_user_and_project_adapter(
        self,
        role_id,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
        context=None,
    ):
        # The parameters for this method must match the parameters for
        # delete_grant so that the notifications.role_assignment decorator
        # will work.

        self.driver.remove_role_from_user_and_project(
            user_id, project_id, role_id
        )
        payload = {'user_id': user_id, 'project_id': project_id}
        notifications.Audit.internal(
            notifications.REMOVE_APP_CREDS_FOR_USER, payload
        )
        self._invalidate_token_cache(
            role_id, group_id, user_id, project_id, domain_id
        )

    def remove_role_from_user_and_project(self, user_id, project_id, role_id):
        self._remove_role_from_user_and_project_adapter(
            role_id, user_id=user_id, project_id=project_id
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def _invalidate_token_cache(
        self, role_id, group_id, user_id, project_id, domain_id
    ):
        if group_id:
            actor_type = 'group'
            actor_id = group_id
        elif user_id:
            actor_type = 'user'
            actor_id = user_id

        if domain_id:
            target_type = 'domain'
            target_id = domain_id
        elif project_id:
            target_type = 'project'
            target_id = project_id

        reason = (
            f'Invalidating the token cache because role {role_id} was '
            f'removed from {actor_type} {actor_id} on {target_type} '
            f'{target_id}.'
        )
        notifications.invalidate_token_cache_notification(reason)

    @notifications.role_assignment('created')
    def create_grant(
        self,
        role_id,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
        initiator=None,
    ):
        role = PROVIDERS.role_api.get_role(role_id)
        if domain_id:
            PROVIDERS.resource_api.get_domain(domain_id)
        if project_id:
            project = PROVIDERS.resource_api.get_project(project_id)

            # For domain specific roles, the domain of the project
            # and role must match
            if role['domain_id'] and project['domain_id'] != role['domain_id']:
                raise exception.DomainSpecificRoleMismatch(
                    role_id=role_id, project_id=project_id
                )

        self.driver.create_grant(
            role_id,
            user_id=user_id,
            group_id=group_id,
            domain_id=domain_id,
            project_id=project_id,
            inherited_to_projects=inherited_to_projects,
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def get_grant(
        self,
        role_id,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
    ):
        role_ref = PROVIDERS.role_api.get_role(role_id)
        if domain_id:
            PROVIDERS.resource_api.get_domain(domain_id)
        if project_id:
            PROVIDERS.resource_api.get_project(project_id)
        self.check_grant_role_id(
            role_id,
            user_id=user_id,
            group_id=group_id,
            domain_id=domain_id,
            project_id=project_id,
            inherited_to_projects=inherited_to_projects,
        )
        return role_ref

    def list_grants(
        self,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
    ):
        if domain_id:
            PROVIDERS.resource_api.get_domain(domain_id)
        if project_id:
            PROVIDERS.resource_api.get_project(project_id)
        grant_ids = self.list_grant_role_ids(
            user_id=user_id,
            group_id=group_id,
            domain_id=domain_id,
            project_id=project_id,
            inherited_to_projects=inherited_to_projects,
        )
        return PROVIDERS.role_api.list_roles_from_ids(grant_ids)

    @notifications.role_assignment('deleted')
    def delete_grant(
        self,
        role_id,
        user_id=None,
        group_id=None,
        domain_id=None,
        project_id=None,
        inherited_to_projects=False,
        initiator=None,
    ):
        # check if role exist before any processing
        PROVIDERS.role_api.get_role(role_id)

        if group_id is None:
            # check if role exists on the user before revoke
            self.check_grant_role_id(
                role_id,
                user_id=user_id,
                group_id=None,
                domain_id=domain_id,
                project_id=project_id,
                inherited_to_projects=inherited_to_projects,
            )
            self._invalidate_token_cache(
                role_id, group_id, user_id, project_id, domain_id
            )
        else:
            try:
                # check if role exists on the group before revoke
                self.check_grant_role_id(
                    role_id,
                    user_id=None,
                    group_id=group_id,
                    domain_id=domain_id,
                    project_id=project_id,
                    inherited_to_projects=inherited_to_projects,
                )
                if CONF.token.revoke_by_id:
                    self._invalidate_token_cache(
                        role_id, group_id, user_id, project_id, domain_id
                    )
            except exception.GroupNotFound:
                LOG.debug(
                    'Group %s not found, no tokens to invalidate.', group_id
                )

        if domain_id:
            PROVIDERS.resource_api.get_domain(domain_id)
        if project_id:
            PROVIDERS.resource_api.get_project(project_id)
        self.driver.delete_grant(
            role_id,
            user_id=user_id,
            group_id=group_id,
            domain_id=domain_id,
            project_id=project_id,
            inherited_to_projects=inherited_to_projects,
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    # The methods _expand_indirect_assignment, _list_direct_role_assignments
    # and _list_effective_role_assignments below are only used on
    # list_role_assignments, but they are not in its scope as nested functions
    # since it would significantly increase McCabe complexity, that should be
    # kept as it is in order to detect unnecessarily complex code, which is not
    # this case.

    def _expand_indirect_assignment(
        self,
        ref,
        user_id=None,
        project_id=None,
        subtree_ids=None,
        expand_groups=True,
    ):
        """Return a list of expanded role assignments.

        This methods is called for each discovered assignment that either needs
        a group assignment expanded into individual user assignments, or needs
        an inherited assignment to be applied to its children.

        In all cases, if either user_id and/or project_id is specified, then we
        filter the result on those values.

        If project_id is specified and subtree_ids is None, then this
        indicates that we are only interested in that one project. If
        subtree_ids is not None, then this is an indicator that any
        inherited assignments need to be expanded down the tree. The
        actual subtree_ids don't need to be used as a filter here, since we
        already ensured only those assignments that could affect them
        were passed to this method.

        If expand_groups is True then we expand groups out to a list of
        assignments, one for each member of that group.

        """

        def create_group_assignment(base_ref, user_id):
            """Create a group assignment from the provided ref."""
            ref = copy.deepcopy(base_ref)

            ref['user_id'] = user_id

            indirect = ref.setdefault('indirect', {})
            indirect['group_id'] = ref.pop('group_id')

            return ref

        def expand_group_assignment(ref, user_id):
            """Expand group role assignment.

            For any group role assignment on a target, it is replaced by a list
            of role assignments containing one for each user of that group on
            that target.

            An example of accepted ref is::

            {
                'group_id': group_id,
                'project_id': project_id,
                'role_id': role_id
            }

            Once expanded, it should be returned as a list of entities like the
            one below, one for each user_id in the provided group_id.

            ::

            {
                'user_id': user_id,
                'project_id': project_id,
                'role_id': role_id,
                'indirect' : {
                    'group_id': group_id
                }
            }

            Returned list will be formatted by the Controller, which will
            deduce a role assignment came from group membership if it has both
            'user_id' in the main body of the dict and 'group_id' in indirect
            subdict.

            """
            if user_id:
                return [create_group_assignment(ref, user_id=user_id)]

            # Note(prashkre): Try to get the users in a group,
            # if a group wasn't found in the backend, users are set
            # as empty list.
            try:
                users = PROVIDERS.identity_api.list_users_in_group(
                    ref['group_id']
                )
            except exception.GroupNotFound:
                LOG.warning(
                    'Group %(group)s was not found but still has role '
                    'assignments.',
                    {'group': ref['group_id']},
                )
                users = []

            return [
                create_group_assignment(ref, user_id=m['id']) for m in users
            ]

        def expand_inherited_assignment(
            ref, user_id, project_id, subtree_ids, expand_groups
        ):
            """Expand inherited role assignments.

            If expand_groups is True and this is a group role assignment on a
            target, replace it by a list of role assignments containing one for
            each user of that group, on every project under that target. If
            expand_groups is False, then return a group assignment on an
            inherited target.

            If this is a user role assignment on a specific target (i.e.
            project_id is specified, but subtree_ids is None) then simply
            format this as a single assignment (since we are effectively
            filtering on project_id). If however, project_id is None or
            subtree_ids is not None, then replace this one assignment with a
            list of role assignments for that user on every project under
            that target.

            An example of accepted ref is::

            {
                'group_id': group_id,
                'project_id': parent_id,
                'role_id': role_id,
                'inherited_to_projects': 'projects'
            }

            Once expanded, it should be returned as a list of entities like the
            one below, one for each user_id in the provided group_id and
            for each subproject_id in the project_id subtree.

            ::

            {
                'user_id': user_id,
                'project_id': subproject_id,
                'role_id': role_id,
                'indirect' : {
                    'group_id': group_id,
                    'project_id': parent_id
                }
            }

            Returned list will be formatted by the Controller, which will
            deduce a role assignment came from group membership if it has both
            'user_id' in the main body of the dict and 'group_id' in the
            'indirect' subdict, as well as it is possible to deduce if it has
            come from inheritance if it contains both a 'project_id' in the
            main body of the dict and 'parent_id' in the 'indirect' subdict.

            """

            def create_inherited_assignment(base_ref, project_id):
                """Create a project assignment from the provided ref.

                base_ref can either be a project or domain inherited
                assignment ref.

                """
                ref = copy.deepcopy(base_ref)

                indirect = ref.setdefault('indirect', {})
                if ref.get('project_id'):
                    indirect['project_id'] = ref.pop('project_id')
                else:
                    indirect['domain_id'] = ref.pop('domain_id')

                ref['project_id'] = project_id
                ref.pop('inherited_to_projects')

                return ref

            # Define expanded project list to which to apply this assignment
            if project_id:
                # Since ref is an inherited assignment and we are filtering by
                # project(s), we are only going to apply the assignment to the
                # relevant project(s)
                project_ids = [project_id]
                if subtree_ids:
                    project_ids += subtree_ids
                    # If this is a domain inherited assignment, then we know
                    # that all the project_ids will get this assignment. If
                    # it's a project inherited assignment, and the assignment
                    # point is an ancestor of project_id, then we know that
                    # again all the project_ids will get the assignment.  If,
                    # however, the assignment point is within the subtree,
                    # then only a partial tree will get the assignment.
                    resource_api = PROVIDERS.resource_api
                    if ref.get('project_id'):
                        if ref['project_id'] in project_ids:
                            project_ids = [
                                x['id']
                                for x in resource_api.list_projects_in_subtree(
                                    ref['project_id']
                                )
                            ]
            elif ref.get('domain_id'):
                # A domain inherited assignment, so apply it to all projects
                # in this domain
                project_ids = [
                    x['id']
                    for x in PROVIDERS.resource_api.list_projects_in_domain(
                        ref['domain_id']
                    )
                ]
            else:
                # It must be a project assignment, so apply it to its subtree
                project_ids = [
                    x['id']
                    for x in PROVIDERS.resource_api.list_projects_in_subtree(
                        ref['project_id']
                    )
                ]

            new_refs = []
            if 'group_id' in ref:
                if expand_groups:
                    # Expand role assignment to all group members on any
                    # inherited target of any of the projects
                    for ref in expand_group_assignment(ref, user_id):
                        new_refs += [
                            create_inherited_assignment(ref, proj_id)
                            for proj_id in project_ids
                        ]
                else:
                    # Just place the group assignment on any inherited target
                    # of any of the projects
                    new_refs += [
                        create_inherited_assignment(ref, proj_id)
                        for proj_id in project_ids
                    ]
            else:
                # Expand role assignment for all projects
                new_refs += [
                    create_inherited_assignment(ref, proj_id)
                    for proj_id in project_ids
                ]

            return new_refs

        if ref.get('inherited_to_projects') == 'projects':
            return expand_inherited_assignment(
                ref, user_id, project_id, subtree_ids, expand_groups
            )
        elif 'group_id' in ref and expand_groups:
            return expand_group_assignment(ref, user_id)
        return [ref]

    def add_implied_roles(self, role_refs):
        """Expand out implied roles.

        The role_refs passed in have had all inheritance and group assignments
        expanded out. We now need to look at the role_id in each ref and see
        if it is a prior role for some implied roles. If it is, then we need to
        duplicate that ref, one for each implied role. We store the prior role
        in the indirect dict that is part of such a duplicated ref, so that a
        caller can determine where the assignment came from.

        """

        def _make_implied_ref_copy(prior_ref, implied_role_id):
            # Create a ref for an implied role from the ref of a prior role,
            # setting the new role_id to be the implied role and the indirect
            # role_id to be the prior role
            implied_ref = copy.deepcopy(prior_ref)
            implied_ref['role_id'] = implied_role_id
            indirect = implied_ref.setdefault('indirect', {})
            indirect['role_id'] = prior_ref['role_id']
            return implied_ref

        try:
            implied_roles_cache = {}
            role_refs_to_check = list(role_refs)
            ref_results = list(role_refs)
            checked_role_refs = []
            while role_refs_to_check:
                next_ref = role_refs_to_check.pop()
                checked_role_refs.append(next_ref)
                next_role_id = next_ref['role_id']
                if next_role_id in implied_roles_cache:
                    implied_roles = implied_roles_cache[next_role_id]
                else:
                    implied_roles = PROVIDERS.role_api.list_implied_roles(
                        next_role_id
                    )
                    implied_roles_cache[next_role_id] = implied_roles
                for implied_role in implied_roles:
                    implied_ref = _make_implied_ref_copy(
                        next_ref, implied_role['implied_role_id']
                    )
                    if implied_ref in checked_role_refs:
                        # Avoid traversing a cycle
                        continue
                    else:
                        ref_results.append(implied_ref)
                        role_refs_to_check.append(implied_ref)
        except exception.NotImplemented:
            LOG.error('Role driver does not support implied roles.')

        return ref_results

    def _filter_by_role_id(self, role_id, ref_results):
        # if we arrive here, we need to filer by role_id.
        filter_results = []
        for ref in ref_results:
            if ref['role_id'] == role_id:
                filter_results.append(ref)
        return filter_results

    def _strip_domain_roles(self, role_refs):
        """Post process assignment list for domain roles.

        Domain roles are only designed to do the job of inferring other roles
        and since that has been done before this method is called, we need to
        remove any assignments that include a domain role.

        """

        def _role_is_global(role_id):
            ref = PROVIDERS.role_api.get_role(role_id)
            return ref['domain_id'] is None

        filter_results = []
        for ref in role_refs:
            if _role_is_global(ref['role_id']):
                filter_results.append(ref)
        return filter_results

    def _list_effective_role_assignments(
        self,
        role_id,
        user_id,
        group_id,
        domain_id,
        project_id,
        subtree_ids,
        inherited,
        source_from_group_ids,
        strip_domain_roles,
    ):
        """List role assignments in effective mode.

        When using effective mode, besides the direct assignments, the indirect
        ones that come from grouping or inheritance are retrieved and will then
        be expanded.

        The resulting list of assignments will be filtered by the provided
        parameters. If subtree_ids is not None, then we also want to include
        all subtree_ids in the filter as well. Since we are in effective mode,
        group can never act as a filter (since group assignments are expanded
        into user roles) and domain can only be filter if we want non-inherited
        assignments, since domains can't inherit assignments.

        The goal of this method is to only ask the driver for those
        assignments as could effect the result based on the parameter filters
        specified, hence avoiding retrieving a huge list.

        """

        def list_role_assignments_for_actor(
            role_id,
            inherited,
            user_id=None,
            group_ids=None,
            project_id=None,
            subtree_ids=None,
            domain_id=None,
        ):
            """List role assignments for actor on target.

            List direct and indirect assignments for an actor, optionally
            for a given target (i.e. projects or domain).

            :param role_id: List for a specific role, can be None meaning all
                            roles
            :param inherited: Indicates whether inherited assignments or only
                              direct assignments are required.  If None, then
                              both are required.
            :param user_id: If not None, list only assignments that affect this
                            user.
            :param group_ids: A list of groups required. Only one of user_id
                              and group_ids can be specified
            :param project_id: If specified, only include those assignments
                               that affect at least this project, with
                               additionally any projects specified in
                               subtree_ids
            :param subtree_ids: The list of projects in the subtree. If
                                specified, also include those assignments that
                                affect these projects. These projects are
                                guaranteed to be in the same domain as the
                                project specified in project_id. subtree_ids
                                can only be specified if project_id has also
                                been specified.
            :param domain_id: If specified, only include those assignments
                              that affect this domain - by definition this will
                              not include any inherited assignments

            :returns: List of assignments matching the criteria. Any inherited
                      or group assignments that could affect the resulting
                      response are included.

            """
            project_ids_of_interest = None
            if project_id:
                if subtree_ids:
                    project_ids_of_interest = subtree_ids + [project_id]
                else:
                    project_ids_of_interest = [project_id]

            # List direct project role assignments
            non_inherited_refs = []
            if inherited is False or inherited is None:
                # Get non inherited assignments
                non_inherited_refs = self.driver.list_role_assignments(
                    role_id=role_id,
                    domain_id=domain_id,
                    project_ids=project_ids_of_interest,
                    user_id=user_id,
                    group_ids=group_ids,
                    inherited_to_projects=False,
                )

            inherited_refs = []
            if inherited is True or inherited is None:
                # Get inherited assignments
                if project_id:
                    # The project and any subtree are guaranteed to be owned by
                    # the same domain, so since we are filtering by these
                    # specific projects, then we can only get inherited
                    # assignments from their common domain or from any of
                    # their parents projects.

                    # List inherited assignments from the project's domain
                    proj_domain_id = PROVIDERS.resource_api.get_project(
                        project_id
                    )['domain_id']
                    inherited_refs += self.driver.list_role_assignments(
                        role_id=role_id,
                        domain_id=proj_domain_id,
                        user_id=user_id,
                        group_ids=group_ids,
                        inherited_to_projects=True,
                    )

                    # For inherited assignments from projects, since we know
                    # they are from the same tree the only places these can
                    # come from are from parents of the main project or
                    # inherited assignments on the project or subtree itself.
                    source_ids = [
                        project['id']
                        for project in PROVIDERS.resource_api.list_project_parents(
                            project_id
                        )
                    ]
                    if subtree_ids:
                        source_ids += project_ids_of_interest
                    if source_ids:
                        inherited_refs += self.driver.list_role_assignments(
                            role_id=role_id,
                            project_ids=source_ids,
                            user_id=user_id,
                            group_ids=group_ids,
                            inherited_to_projects=True,
                        )
                else:
                    # List inherited assignments without filtering by target
                    inherited_refs = self.driver.list_role_assignments(
                        role_id=role_id,
                        user_id=user_id,
                        group_ids=group_ids,
                        inherited_to_projects=True,
                    )

            return non_inherited_refs + inherited_refs

        # If filtering by group or inherited domain assignment the list is
        # guaranteed to be empty
        if group_id or (domain_id and inherited):
            return []

        if user_id and source_from_group_ids:
            # You can't do both - and since source_from_group_ids is only used
            # internally, this must be a coding error by the caller.
            msg = _(
                'Cannot list assignments sourced from groups and filtered '
                'by user ID.'
            )
            raise exception.UnexpectedError(msg)

        # If filtering by domain, then only non-inherited assignments are
        # relevant, since domains don't inherit assignments
        inherited = False if domain_id else inherited

        # List user or explicit group assignments.
        # Due to the need to expand implied roles, this call will skip
        # filtering by role_id and instead return the whole set of roles.
        # Matching on the specified role is performed at the end.
        direct_refs = list_role_assignments_for_actor(
            role_id=None,
            user_id=user_id,
            group_ids=source_from_group_ids,
            project_id=project_id,
            subtree_ids=subtree_ids,
            domain_id=domain_id,
            inherited=inherited,
        )

        # And those from the user's groups, so long as we are not restricting
        # to a set of source groups (in which case we already got those
        # assignments in the direct listing above).
        group_refs = []
        if not source_from_group_ids and user_id:
            group_ids = self._get_group_ids_for_user_id(user_id)
            if group_ids:
                group_refs = list_role_assignments_for_actor(
                    role_id=None,
                    project_id=project_id,
                    subtree_ids=subtree_ids,
                    group_ids=group_ids,
                    domain_id=domain_id,
                    inherited=inherited,
                )

        # Expand grouping and inheritance on retrieved role assignments
        refs = []
        expand_groups = source_from_group_ids is None
        for ref in direct_refs + group_refs:
            refs += self._expand_indirect_assignment(
                ref, user_id, project_id, subtree_ids, expand_groups
            )

        refs = self.add_implied_roles(refs)
        if strip_domain_roles:
            refs = self._strip_domain_roles(refs)
        if role_id:
            refs = self._filter_by_role_id(role_id, refs)

        return refs

    def _list_direct_role_assignments(
        self,
        role_id,
        user_id,
        group_id,
        system,
        domain_id,
        project_id,
        subtree_ids,
        inherited,
    ):
        """List role assignments without applying expansion.

        Returns a list of direct role assignments, where their attributes match
        the provided filters. If subtree_ids is not None, then we also want to
        include all subtree_ids in the filter as well.

        """
        group_ids = [group_id] if group_id else None
        project_ids_of_interest = None
        if project_id:
            if subtree_ids:
                project_ids_of_interest = subtree_ids + [project_id]
            else:
                project_ids_of_interest = [project_id]

        project_and_domain_assignments = []
        if not system:
            project_and_domain_assignments = self.driver.list_role_assignments(
                role_id=role_id,
                user_id=user_id,
                group_ids=group_ids,
                domain_id=domain_id,
                project_ids=project_ids_of_interest,
                inherited_to_projects=inherited,
            )

        system_assignments = []
        if system or (not project_id and not domain_id and not system):
            if user_id:
                assignments = self.list_system_grants_for_user(user_id)
                for assignment in assignments:
                    system_assignments.append(
                        {
                            'system': {'all': True},
                            'user_id': user_id,
                            'role_id': assignment['id'],
                        }
                    )
            elif group_id:
                assignments = self.list_system_grants_for_group(group_id)
                for assignment in assignments:
                    system_assignments.append(
                        {
                            'system': {'all': True},
                            'group_id': group_id,
                            'role_id': assignment['id'],
                        }
                    )
            else:
                assignments = self.list_all_system_grants()
                for assignment in assignments:
                    a = {}
                    if assignment['type'] == self._GROUP_SYSTEM:
                        a['group_id'] = assignment['actor_id']
                    elif assignment['type'] == self._USER_SYSTEM:
                        a['user_id'] = assignment['actor_id']
                    a['role_id'] = assignment['role_id']
                    a['system'] = {'all': True}
                    system_assignments.append(a)

            if role_id:
                system_assignments = [
                    sa for sa in system_assignments if role_id == sa['role_id']
                ]

        assignments = []
        for assignment in itertools.chain(
            project_and_domain_assignments, system_assignments
        ):
            assignments.append(assignment)

        return assignments

    @MEMOIZE_COMPUTED_ASSIGNMENTS
    def list_role_assignments(
        self,
        role_id=None,
        user_id=None,
        group_id=None,
        system=None,
        domain_id=None,
        project_id=None,
        include_subtree=False,
        inherited=None,
        effective=None,
        include_names=False,
        source_from_group_ids=None,
        strip_domain_roles=True,
    ):
        """List role assignments, honoring effective mode and provided filters.

        Returns a list of role assignments, where their attributes match the
        provided filters (role_id, user_id, group_id, domain_id, project_id and
        inherited). If include_subtree is True, then assignments on all
        descendants of the project specified by project_id are also included.
        The inherited filter defaults to None, meaning to get both
        non-inherited and inherited role assignments.

        If effective mode is specified, this means that rather than simply
        return the assignments that match the filters, any group or
        inheritance assignments will be expanded. Group assignments will
        become assignments for all the users in that group, and inherited
        assignments will be shown on the projects below the assignment point.
        Think of effective mode as being the list of assignments that actually
        affect a user, for example the roles that would be placed in a token.

        If include_names is set to true the entities' names are returned
        in addition to their id's.

        source_from_group_ids is a list of group IDs and, if specified, then
        only those assignments that are derived from membership of these groups
        are considered, and any such assignments will not be expanded into
        their user membership assignments. This is different to a group filter
        of the resulting list, instead being a restriction on which assignments
        should be considered before expansion of inheritance. This option is
        only used internally (i.e. it is not exposed at the API level) and is
        only supported in effective mode (since in regular mode there is no
        difference between this and a group filter, other than it is a list of
        groups).

        In effective mode, any domain specific roles are usually stripped from
        the returned assignments (since such roles are not placed in tokens).
        This stripping can be disabled by specifying strip_domain_roles=False,
        which is useful for internal calls like trusts which need to examine
        the full set of roles.
        """
        subtree_ids = None
        if project_id and include_subtree:
            subtree_ids = [
                x['id']
                for x in PROVIDERS.resource_api.list_projects_in_subtree(
                    project_id
                )
            ]

        if system != 'all':
            system = None

        if effective:
            role_assignments = self._list_effective_role_assignments(
                role_id,
                user_id,
                group_id,
                domain_id,
                project_id,
                subtree_ids,
                inherited,
                source_from_group_ids,
                strip_domain_roles,
            )
        else:
            role_assignments = self._list_direct_role_assignments(
                role_id,
                user_id,
                group_id,
                system,
                domain_id,
                project_id,
                subtree_ids,
                inherited,
            )

        if include_names:
            return self._get_names_from_role_assignments(role_assignments)
        return role_assignments

    def _get_names_from_role_assignments(self, role_assignments):
        role_assign_list = []

        for role_asgmt in role_assignments:
            new_assign = copy.deepcopy(role_asgmt)
            for key, value in role_asgmt.items():
                if key == 'domain_id':
                    _domain = PROVIDERS.resource_api.get_domain(value)
                    new_assign['domain_name'] = _domain['name']
                elif key == 'user_id':
                    try:
                        # Note(knikolla): Try to get the user, otherwise
                        # if the user wasn't found in the backend
                        # use empty values.
                        _user = PROVIDERS.identity_api.get_user(value)
                    except exception.UserNotFound:
                        msg = (
                            'User %(user)s not found in the'
                            ' backend but still has role assignments.'
                        )
                        LOG.warning(msg, {'user': value})
                        new_assign['user_name'] = ''
                        new_assign['user_domain_id'] = ''
                        new_assign['user_domain_name'] = ''
                    else:
                        new_assign['user_name'] = _user['name']
                        new_assign['user_domain_id'] = _user['domain_id']
                        new_assign['user_domain_name'] = (
                            PROVIDERS.resource_api.get_domain(
                                _user['domain_id']
                            )['name']
                        )
                elif key == 'group_id':
                    try:
                        # Note(knikolla): Try to get the group, otherwise
                        # if the group wasn't found in the backend
                        # use empty values.
                        _group = PROVIDERS.identity_api.get_group(value)
                    except exception.GroupNotFound:
                        msg = (
                            'Group %(group)s not found in the'
                            ' backend but still has role assignments.'
                        )
                        LOG.warning(msg, {'group': value})
                        new_assign['group_name'] = ''
                        new_assign['group_domain_id'] = ''
                        new_assign['group_domain_name'] = ''
                    else:
                        new_assign['group_name'] = _group['name']
                        new_assign['group_domain_id'] = _group['domain_id']
                        new_assign['group_domain_name'] = (
                            PROVIDERS.resource_api.get_domain(
                                _group['domain_id']
                            )['name']
                        )
                elif key == 'project_id':
                    _project = PROVIDERS.resource_api.get_project(value)
                    new_assign['project_name'] = _project['name']
                    new_assign['project_domain_id'] = _project['domain_id']
                    new_assign['project_domain_name'] = (
                        PROVIDERS.resource_api.get_domain(
                            _project['domain_id']
                        )['name']
                    )
                elif key == 'role_id':
                    _role = PROVIDERS.role_api.get_role(value)
                    new_assign['role_name'] = _role['name']
                    if _role['domain_id'] is not None:
                        new_assign['role_domain_id'] = _role['domain_id']
                        new_assign['role_domain_name'] = (
                            PROVIDERS.resource_api.get_domain(
                                _role['domain_id']
                            )['name']
                        )
            role_assign_list.append(new_assign)
        return role_assign_list

    def delete_group_assignments(self, group_id):
        # FIXME(lbragstad): This should be refactored in the Rocky release so
        # that we can pass the group_id to the system assignment backend like
        # we do with the project and domain assignment backend. Holding off on
        # this because it will require an interface change to the backend,
        # making it harder to backport for Queens RC.
        self.driver.delete_group_assignments(group_id)
        system_assignments = self.list_system_grants_for_group(group_id)
        for assignment in system_assignments:
            self.delete_system_grant_for_group(group_id, assignment['id'])
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def delete_user_assignments(self, user_id):
        # FIXME(lbragstad): This should be refactored in the Rocky release so
        # that we can pass the user_id to the system assignment backend like we
        # do with the project and domain assignment backend. Holding off on
        # this because it will require an interface change to the backend,
        # making it harder to backport for Queens RC.
        self.driver.delete_user_assignments(user_id)
        system_assignments = self.list_system_grants_for_user(user_id)
        for assignment in system_assignments:
            self.delete_system_grant_for_user(user_id, assignment['id'])
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def check_system_grant_for_user(self, user_id, role_id):
        """Check if a user has a specific role on the system.

        :param user_id: the ID of the user in the assignment
        :param role_id: the ID of the system role in the assignment

        :raises keystone.exception.RoleAssignmentNotFound: if the user doesn't
            have a role assignment matching the role_id on the system

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        inherited = False
        return self.driver.check_system_grant(
            role_id, user_id, target_id, inherited
        )

    def list_system_grants_for_user(self, user_id):
        """Return a list of roles the user has on the system.

        :param user_id: the ID of the user

        :returns: a list of role assignments the user has system-wide

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        assignment_type = self._USER_SYSTEM
        grants = self.driver.list_system_grants(
            user_id, target_id, assignment_type
        )
        grant_ids = []
        for grant in grants:
            grant_ids.append(grant['role_id'])

        return PROVIDERS.role_api.list_roles_from_ids(grant_ids)

    def create_system_grant_for_user(self, user_id, role_id):
        """Grant a user a role on the system.

        :param user_id: the ID of the user
        :param role_id: the ID of the role to grant on the system

        """
        role = PROVIDERS.role_api.get_role(role_id)
        if role.get('domain_id'):
            raise exception.ValidationError(
                f'Role {role_id} is a domain-specific role. Unable to use '
                'a domain-specific role in a system assignment.'
            )
        target_id = self._SYSTEM_SCOPE_TOKEN
        assignment_type = self._USER_SYSTEM
        inherited = False
        self.driver.create_system_grant(
            role_id, user_id, target_id, assignment_type, inherited
        )

    def delete_system_grant_for_user(self, user_id, role_id):
        """Remove a system grant from a user.

        :param user_id: the ID of the user
        :param role_id: the ID of the role to remove from the user on the
                        system

        :raises keystone.exception.RoleAssignmentNotFound: if the user doesn't
            have a role assignment with role_id on the system

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        inherited = False
        self.driver.delete_system_grant(role_id, user_id, target_id, inherited)
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def check_system_grant_for_group(self, group_id, role_id):
        """Check if a group has a specific role on the system.

        :param group_id: the ID of the group in the assignment
        :param role_id: the ID of the system role in the assignment

        :raises keystone.exception.RoleAssignmentNotFound: if the group doesn't
            have a role assignment matching the role_id on the system

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        inherited = False
        return self.driver.check_system_grant(
            role_id, group_id, target_id, inherited
        )

    def list_system_grants_for_group(self, group_id):
        """Return a list of roles the group has on the system.

        :param group_id: the ID of the group

        :returns: a list of role assignments the group has system-wide

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        assignment_type = self._GROUP_SYSTEM
        grants = self.driver.list_system_grants(
            group_id, target_id, assignment_type
        )
        grant_ids = []
        for grant in grants:
            grant_ids.append(grant['role_id'])

        return PROVIDERS.role_api.list_roles_from_ids(grant_ids)

    def create_system_grant_for_group(self, group_id, role_id):
        """Grant a group a role on the system.

        :param group_id: the ID of the group
        :param role_id: the ID of the role to grant on the system

        """
        role = PROVIDERS.role_api.get_role(role_id)
        if role.get('domain_id'):
            raise exception.ValidationError(
                f'Role {role_id} is a domain-specific role. Unable to use '
                'a domain-specific role in a system assignment.'
            )
        target_id = self._SYSTEM_SCOPE_TOKEN
        assignment_type = self._GROUP_SYSTEM
        inherited = False
        self.driver.create_system_grant(
            role_id, group_id, target_id, assignment_type, inherited
        )

    def delete_system_grant_for_group(self, group_id, role_id):
        """Remove a system grant from a group.

        :param group_id: the ID of the group
        :param role_id: the ID of the role to remove from the group on the
                        system

        :raises keystone.exception.RoleAssignmentNotFound: if the group doesn't
            have a role assignment with role_id on the system

        """
        target_id = self._SYSTEM_SCOPE_TOKEN
        inherited = False
        self.driver.delete_system_grant(
            role_id, group_id, target_id, inherited
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    def list_all_system_grants(self):
        """Return a list of all system grants."""
        actor_id = None
        target_id = self._SYSTEM_SCOPE_TOKEN
        assignment_type = None
        return self.driver.list_system_grants(
            actor_id, target_id, assignment_type
        )


class RoleManager(manager.Manager):
    """Default pivot point for the Role backend."""

    driver_namespace = 'keystone.role'
    _provides_api = 'role_api'

    _ROLE = 'role'

    def __init__(self):
        # If there is a specific driver specified for role, then use it.
        # Otherwise retrieve the driver type from the assignment driver.
        role_driver = CONF.role.driver

        if role_driver is None:
            # Explicitly load the assignment manager object
            assignment_driver = CONF.assignment.driver
            assignment_manager_obj = manager.load_driver(
                Manager.driver_namespace, assignment_driver
            )
            role_driver = assignment_manager_obj.default_role_driver()

        super().__init__(role_driver)

    @MEMOIZE
    def get_role(self, role_id):
        return self.driver.get_role(role_id)

    def get_unique_role_by_name(self, role_name, hints=None):
        if not hints:
            hints = driver_hints.Hints()
        hints.add_filter("name", role_name, case_sensitive=True)
        found_roles = PROVIDERS.role_api.list_roles(hints)
        if not found_roles:
            raise exception.RoleNotFound(
                _("Role %s is not defined") % role_name
            )
        elif len(found_roles) == 1:
            return {'id': found_roles[0]['id']}
        else:
            raise exception.AmbiguityError(resource='role', name=role_name)

    def create_role(self, role_id, role, initiator=None):
        # Shallow copy to help mitigate in-line changes that might impact
        # testing. This mirrors create_user, specifically relevant for
        # resource options.
        role = role.copy()
        ret = self.driver.create_role(role_id, role)
        notifications.Audit.created(self._ROLE, role_id, initiator)
        if MEMOIZE.should_cache(ret):
            self.get_role.set(ret, self, role_id)
        return ret

    @manager.response_truncated
    def list_roles(self, hints=None):
        return self.driver.list_roles(hints or driver_hints.Hints())

    def _is_immutable(self, role):
        return role['options'].get(ro_opt.IMMUTABLE_OPT.option_name, False)

    def update_role(self, role_id, role, initiator=None):
        original_role = self.driver.get_role(role_id)
        # Prevent the update of immutable set roles unless the update is
        # exclusively used for
        ro_opt.check_immutable_update(
            original_resource_ref=original_role,
            new_resource_ref=role,
            type='role',
            resource_id=role_id,
        )

        if (
            'domain_id' in role
            and role['domain_id'] != original_role['domain_id']
        ):
            raise exception.ValidationError(
                message=_('Update of `domain_id` is not allowed.')
            )

        ret = self.driver.update_role(role_id, role)
        notifications.Audit.updated(self._ROLE, role_id, initiator)
        self.get_role.invalidate(self, role_id)
        return ret

    def delete_role(self, role_id, initiator=None):
        role = self.driver.get_role(role_id)
        # Prevent deletion of immutable roles.
        ro_opt.check_immutable_delete(
            resource_ref=role, resource_type='role', resource_id=role_id
        )
        PROVIDERS.assignment_api._send_app_cred_notification_for_role_removal(
            role_id
        )
        PROVIDERS.assignment_api.delete_role_assignments(role_id)
        self.driver.delete_role(role_id)
        notifications.Audit.deleted(self._ROLE, role_id, initiator)
        self.get_role.invalidate(self, role_id)
        reason = (
            f'Invalidating the token cache because role {role_id} has been '
            'removed. Role assignments for users will be recalculated and '
            'enforced accordingly the next time they authenticate or validate '
            'a token'
        )
        notifications.invalidate_token_cache_notification(reason)
        COMPUTED_ASSIGNMENTS_REGION.invalidate()

    # TODO(ayoung): Add notification
    def create_implied_role(self, prior_role_id, implied_role_id):
        implied_role = self.driver.get_role(implied_role_id)
        prior_role = self.driver.get_role(prior_role_id)
        if implied_role['name'] in CONF.assignment.prohibited_implied_role:
            raise exception.InvalidImpliedRole(role_id=implied_role_id)
        if prior_role['domain_id'] is None and implied_role['domain_id']:
            msg = _('Global role cannot imply a domain-specific role')
            raise exception.InvalidImpliedRole(msg, role_id=implied_role_id)
        response = self.driver.create_implied_role(
            prior_role_id, implied_role_id
        )
        COMPUTED_ASSIGNMENTS_REGION.invalidate()
        return response

    def delete_implied_role(self, prior_role_id, implied_role_id):
        self.driver.delete_implied_role(prior_role_id, implied_role_id)
        COMPUTED_ASSIGNMENTS_REGION.invalidate()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fix incorrect constraints.

Revision ID: b4f8b3f584e0
Revises: 29e87d24a316
Create Date: 2022-08-02 12:23:25.520570
"""

from alembic import op

# revision identifiers, used by Alembic.
revision = 'b4f8b3f584e0'
down_revision = '29e87d24a316'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('trust', schema=None) as batch_op:
        batch_op.create_unique_constraint(
            'duplicate_trust_constraint',
            [
                'trustor_user_id',
                'trustee_user_id',
                'project_id',
                'impersonation',
                'expires_at',
            ],
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any

from keystone.api.validation import parameter_types
from keystone.api.validation import response_types
from keystone.common import validation
from keystone.common.validation import parameter_types as ks_parameter_types

basic_property_id = {
    'type': 'object',
    'properties': {'id': {'type': 'string'}},
    'required': ['id'],
    'additionalProperties': False,
}

saml_create = {
    'type': 'object',
    'properties': {
        'identity': {
            'type': 'object',
            'properties': {
                'token': basic_property_id,
                'methods': {'type': 'array'},
            },
            'required': ['token'],
            'additionalProperties': False,
        },
        'scope': {
            'type': 'object',
            'properties': {'service_provider': basic_property_id},
            'required': ['service_provider'],
            'additionalProperties': False,
        },
    },
    'required': ['identity', 'scope'],
    'additionalProperties': False,
}

_service_provider_properties = {
    # NOTE(rodrigods): The database accepts URLs with 256 as max length,
    # but parameter_types.url uses 225 as max length.
    "auth_url": {
        **parameter_types.url,
        "description": "The URL to authenticate against",
    },
    "sp_url": {
        **parameter_types.url,
        "description": "The service provider's URL",
    },
    "description": {
        "type": ["string", "null"],
        "description": "The description of the service provider",
    },
    "enabled": {
        **parameter_types.boolean,
        "description": "Whether the service provider is enabled or not",
    },
    "relay_state_prefix": {
        "type": ["string", "null"],
        "description": "The prefix of the RelayState SAML attribute",
    },
}

service_provider_schema: dict[str, Any] = {
    "type": "object",
    "description": "A service provider object",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The service provider ID",
        },
        "links": response_types.resource_links,
        **_service_provider_properties,
    },
    "additionalProperties": True,
}

service_provider_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "id": {"type": "string", "description": "The service provider ID"},
        "enabled": {
            **parameter_types.boolean,
            "description": "Whether the service provider is enabled or not",
        },
    },
    "additionalProperties": False,
}

service_provider_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "service_providers": {
            "type": "array",
            "items": service_provider_schema,
            "description": "A list of service provider objects",
        },
        "links": response_types.links,
        "truncated": response_types.truncated,
    },
    "additionalProperties": False,
}

service_provider_response_body: dict[str, Any] = {
    "type": "object",
    "description": "A service provider object",
    "properties": {"service_provider": service_provider_schema},
    "additionalProperties": False,
}

service_provider_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A service provider object",
    "properties": {
        "service_provider": {
            "type": "object",
            "properties": _service_provider_properties,
            "required": ["auth_url", "sp_url"],
            "additionalProperties": False,
        }
    },
    "required": ["service_provider"],
    "additionalProperties": False,
}

service_provider_update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "A service provider object",
    "properties": {
        # NOTE(rodrigods): 'id' is not required since it is passed in the URL
        "service_provider": {
            "type": "object",
            "properties": _service_provider_properties,
            "additionalProperties": False,
            "minProperties": 1,
        }
    },
    "required": ["service_provider"],
    "additionalProperties": False,
}

identity_provider_response_links: dict[str, Any] = {
    "type": "object",
    "properties": {
        "self": {"type": "string", "format": "uri"},
        "protocols": {"type": "string", "format": "uri"},
    },
    "additionalProperties": False,
}

_identity_provider_properties = {
    "enabled": {
        **parameter_types.boolean,
        "description": "Whether the identity provider is enabled or not",
    },
    "description": {
        "type": ["string", "null"],
        "description": "The identity provider description",
    },
    "authorization_ttl": {
        "type": ["integer", "null"],
        "minimum": 0,
        "description": (
            "The length of validity in minutes for group memberships "
            "carried over through mapping and persisted in the database. "
            "If left unset, the default value configured in keystone will "
            "be used, if enabled."
        ),
    },
    "remote_ids": {
        "type": ["array", "null"],
        "description": "List of the unique identity provider's remote IDs",
        "items": {"type": "string"},
        "uniqueItems": True,
    },
}

identity_provider_schema: dict[str, Any] = {
    "type": "object",
    "description": "An identity provider object",
    "properties": {
        "id": {
            "type": "string",
            "readOnly": True,
            "description": "The identity provider ID",
        },
        "links": identity_provider_response_links,
        **_identity_provider_properties,
    },
    "additionalProperties": True,
}

identity_provider_index_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {
        "id": {"type": "string", "description": "The identity provider ID"},
        "enabled": {
            **parameter_types.boolean,
            "description": "Whether the identity provider is enabled or not",
        },
    },
    "additionalProperties": True,
}

identity_provider_index_response_body: dict[str, Any] = {
    "type": "object",
    "properties": {
        "identity_providers": {
            "type": "array",
            "items": identity_provider_schema,
            "description": "A list of identity provider objects",
        },
        "links": response_types.links,
    },
    "additionalProperties": True,
}

identity_provider_request_query: dict[str, Any] = {
    "type": "object",
    "properties": {},
    "additionalProperties": True,
}

identity_provider_response_body: dict[str, Any] = {
    "type": "object",
    "description": "An identity provider object",
    "properties": {"identity_provider": identity_provider_schema},
    "additionalProperties": False,
}

identity_provider_create_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An identity provider object",
    "properties": {
        "identity_provider": {
            "type": "object",
            "properties": {
                **_identity_provider_properties,
                "domain_id": {
                    "type": ["string", "null"],
                    "minLength": 1,
                    "maxLength": 64,
                    "pattern": r"^[a-zA-Z0-9-]+$",
                    "description": (
                        "The ID of a domain that is associated with the "
                        "identity provider. Federated users that "
                        "authenticate with the identity provider will be "
                        "created under the domain specified."
                    ),
                },
            },
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
}

identity_provider_update_request_body: dict[str, Any] = {
    "type": "object",
    "description": "An identity provider object",
    "properties": {
        "identity_provider": {
            "type": "object",
            "properties": _identity_provider_properties,
            "additionalProperties": False,
            "minProperties": 1,
        }
    },
    "required": ["identity_provider"],
}

_remote_id_attribute_properties = {'type': 'string', 'maxLength': 64}

_protocol_properties = {
    'mapping_id': ks_parameter_types.mapping_id_string,
    'remote_id_attribute': _remote_id_attribute_properties,
}

protocol_create = {
    'type': 'object',
    'properties': _protocol_properties,
    'required': ['mapping_id'],
    'additionalProperties': False,
}

protocol_update = {
    'type': 'object',
    'properties': _protocol_properties,
    'minProperties': 1,
    'additionalProperties': False,
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

# Two of the three portions of this check string are specific to domain
# readers. The first catches domain readers who are checking or listing grants
# for users. The second does the same for groups. We have to overload the check
# string to handle both cases because `identity:check_grant` is used to protect
# both user and group grant APIs. If the `identity:check_grant` policy is every
# broken apart, we can write specific check strings that are tailored to either
# users or groups (e.g., `identity:check_group_grant` or
# `identity:check_user_grant`) and prevent overloading like this.
DOMAIN_MATCHES_USER_DOMAIN = 'domain_id:%(target.user.domain_id)s'
DOMAIN_MATCHES_GROUP_DOMAIN = 'domain_id:%(target.group.domain_id)s'
DOMAIN_MATCHES_PROJECT_DOMAIN = 'domain_id:%(target.project.domain_id)s'
DOMAIN_MATCHES_TARGET_DOMAIN = 'domain_id:%(target.domain.id)s'
DOMAIN_MATCHES_ROLE = (
    'domain_id:%(target.role.domain_id)s or None:%(target.role.domain_id)s'
)
GRANTS_DOMAIN_READER = (
    '(role:reader and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:reader and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ') or '
    '(role:reader and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:reader and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ')'
)
SYSTEM_READER_OR_DOMAIN_READER = (
    '(' + base.SYSTEM_READER + ') or '
    '(' + GRANTS_DOMAIN_READER + ') and '
    '(' + DOMAIN_MATCHES_ROLE + ')'
)

SYSTEM_READER_OR_DOMAIN_READER_LIST = (
    '(' + base.SYSTEM_READER + ') or ' + GRANTS_DOMAIN_READER
)

GRANTS_DOMAIN_ADMIN = (
    '(role:admin and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:admin and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ') or '
    '(role:admin and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:admin and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ')'
)

GRANTS_DOMAIN_MANAGER = (
    '(role:manager and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:manager and ' + DOMAIN_MATCHES_USER_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ') or '
    '(role:manager and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_PROJECT_DOMAIN + ') or '
    '(role:manager and ' + DOMAIN_MATCHES_GROUP_DOMAIN + ' and'
    ' ' + DOMAIN_MATCHES_TARGET_DOMAIN + ')'
)

ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER = (
    '(' + base.RULE_ADMIN_REQUIRED + ') or '
    '(' + SYSTEM_READER_OR_DOMAIN_READER + ')'
)

ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER_LIST = (
    '(' + base.RULE_ADMIN_REQUIRED + ') or '
    '(' + SYSTEM_READER_OR_DOMAIN_READER_LIST + ')'
)

ADMIN_OR_DOMAIN_ADMIN_OR_DOMAIN_MANAGER = (
    '(' + base.RULE_ADMIN_REQUIRED + ') or '
    '(' + GRANTS_DOMAIN_ADMIN + ') and '
    '(' + DOMAIN_MATCHES_ROLE + ') or '
    '(' + GRANTS_DOMAIN_MANAGER + ') and '
    'rule:domain_managed_target_role'
)

DEPRECATED_REASON = (
    "The assignment API is now aware of system scope and default roles."
)

deprecated_check_system_grant_for_user = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_system_grant_for_user',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_list_system_grants_for_user = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_system_grants_for_user',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_create_system_grant_for_user = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_system_grant_for_user',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_revoke_system_grant_for_user = policy.DeprecatedRule(
    name=base.IDENTITY % 'revoke_system_grant_for_user',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_check_system_grant_for_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_system_grant_for_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_list_system_grants_for_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_system_grants_for_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_create_system_grant_for_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_system_grant_for_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_revoke_system_grant_for_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'revoke_system_grant_for_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_list_grants = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_grants',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_check_grant = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_grant',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_create_grant = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_grant',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_revoke_grant = policy.DeprecatedRule(
    name=base.IDENTITY % 'revoke_grant',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)


resource_paths = [
    '/projects/{project_id}/users/{user_id}/roles/{role_id}',
    '/projects/{project_id}/groups/{group_id}/roles/{role_id}',
    '/domains/{domain_id}/users/{user_id}/roles/{role_id}',
    '/domains/{domain_id}/groups/{group_id}/roles/{role_id}',
]


resource_paths += [
    '/OS-INHERIT' + path + '/inherited_to_projects' for path in resource_paths
]


collection_paths = [
    '/projects/{project_id}/users/{user_id}/roles',
    '/projects/{project_id}/groups/{group_id}/roles',
    '/domains/{domain_id}/users/{user_id}/roles',
    '/domains/{domain_id}/groups/{group_id}/roles',
]


inherited_collection_paths = [
    (
        '/OS-INHERIT/domains/{domain_id}/groups/{group_id}/roles/'
        'inherited_to_projects'
    ),
    (
        '/OS-INHERIT/domains/{domain_id}/users/{user_id}/roles/'
        'inherited_to_projects'
    ),
]


def list_operations(paths, methods):
    return [
        {'path': '/v3' + path, 'method': method}
        for path in paths
        for method in methods
    ]


# NOTE(samueldmq): Unlike individual resource paths, collection
# paths for the inherited grants do not contain a HEAD API
list_grants_operations = list_operations(
    collection_paths, ['GET', 'HEAD']
) + list_operations(inherited_collection_paths, ['GET'])


grant_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_grant',
        check_str=ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER,
        scope_types=['system', 'domain', 'project'],
        description=(
            'Check a role grant between a target and an actor. A '
            'target can be either a domain or a project. An actor '
            'can be either a user or a group. These terms also apply '
            'to the OS-INHERIT APIs, where grants on the target '
            'are inherited to all projects in the subtree, if '
            'applicable.'
        ),
        operations=list_operations(resource_paths, ['HEAD', 'GET']),
        deprecated_rule=deprecated_check_grant,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_grants',
        check_str=ADMIN_OR_SYSTEM_READER_OR_DOMAIN_READER_LIST,
        scope_types=['system', 'domain', 'project'],
        description=(
            'List roles granted to an actor on a target. A target '
            'can be either a domain or a project. An actor can be '
            'either a user or a group. For the OS-INHERIT APIs, it '
            'is possible to list inherited role grants for actors on '
            'domains, where grants are inherited to all projects '
            'in the specified domain.'
        ),
        operations=list_grants_operations,
        deprecated_rule=deprecated_list_grants,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_grant',
        check_str=ADMIN_OR_DOMAIN_ADMIN_OR_DOMAIN_MANAGER,
        scope_types=['system', 'domain', 'project'],
        description=(
            'Create a role grant between a target and an actor. A '
            'target can be either a domain or a project. An actor '
            'can be either a user or a group. These terms also apply '
            'to the OS-INHERIT APIs, where grants on the target '
            'are inherited to all projects in the subtree, if '
            'applicable.'
        ),
        operations=list_operations(resource_paths, ['PUT']),
        deprecated_rule=deprecated_create_grant,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'revoke_grant',
        check_str=ADMIN_OR_DOMAIN_ADMIN_OR_DOMAIN_MANAGER,
        scope_types=['system', 'domain', 'project'],
        description=(
            'Revoke a role grant between a target and an actor. A '
            'target can be either a domain or a project. An actor '
            'can be either a user or a group. These terms also apply '
            'to the OS-INHERIT APIs, where grants on the target '
            'are inherited to all projects in the subtree, if '
            'applicable. In that case, revoking the role grant in '
            'the target would remove the logical effect of '
            'inheriting it to the target\'s projects subtree.'
        ),
        operations=list_operations(resource_paths, ['DELETE']),
        deprecated_rule=deprecated_revoke_grant,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_system_grants_for_user',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List all grants a specific user has on the system.',
        operations=[
            {
                'path': '/v3/system/users/{user_id}/roles',
                'method': ['HEAD', 'GET'],
            }
        ],
        deprecated_rule=deprecated_list_system_grants_for_user,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_system_grant_for_user',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check if a user has a role on the system.',
        operations=[
            {
                'path': '/v3/system/users/{user_id}/roles/{role_id}',
                'method': ['HEAD', 'GET'],
            }
        ],
        deprecated_rule=deprecated_check_system_grant_for_user,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_system_grant_for_user',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Grant a user a role on the system.',
        operations=[
            {
                'path': '/v3/system/users/{user_id}/roles/{role_id}',
                'method': ['PUT'],
            }
        ],
        deprecated_rule=deprecated_create_system_grant_for_user,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'revoke_system_grant_for_user',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Remove a role from a user on the system.',
        operations=[
            {
                'path': '/v3/system/users/{user_id}/roles/{role_id}',
                'method': ['DELETE'],
            }
        ],
        deprecated_rule=deprecated_revoke_system_grant_for_user,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_system_grants_for_group',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List all grants a specific group has on the system.',
        operations=[
            {
                'path': '/v3/system/groups/{group_id}/roles',
                'method': ['HEAD', 'GET'],
            }
        ],
        deprecated_rule=deprecated_list_system_grants_for_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_system_grant_for_group',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check if a group has a role on the system.',
        operations=[
            {
                'path': '/v3/system/groups/{group_id}/roles/{role_id}',
                'method': ['HEAD', 'GET'],
            }
        ],
        deprecated_rule=deprecated_check_system_grant_for_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_system_grant_for_group',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Grant a group a role on the system.',
        operations=[
            {
                'path': '/v3/system/groups/{group_id}/roles/{role_id}',
                'method': ['PUT'],
            }
        ],
        deprecated_rule=deprecated_create_system_grant_for_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'revoke_system_grant_for_group',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Remove a role from a group on the system.',
        operations=[
            {
                'path': '/v3/system/groups/{group_id}/roles/{role_id}',
                'method': ['DELETE'],
            }
        ],
        deprecated_rule=deprecated_revoke_system_grant_for_group,
    ),
]


def list_rules():
    return grant_policies



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from unittest import mock
import uuid

import fixtures
import freezegun
from oslo_db import exception as db_exception
from oslo_db import options
from oslo_log import log
from oslo_utils import timeutils
import sqlalchemy
from sqlalchemy import exc
from testtools import matchers

from keystone.common import driver_hints
from keystone.common import provider_api
from keystone.common import sql
from keystone.common.sql import core
import keystone.conf
from keystone.credential.providers import fernet as credential_provider
from keystone import exception
from keystone.identity.backends import sql_model as identity_sql
from keystone.resource.backends import base as resource
from keystone.tests import unit
from keystone.tests.unit.assignment import test_backends as assignment_tests
from keystone.tests.unit.catalog import test_backends as catalog_tests
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.identity import test_backends as identity_tests
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database
from keystone.tests.unit.limit import test_backends as limit_tests
from keystone.tests.unit.policy import test_backends as policy_tests
from keystone.tests.unit.resource import test_backends as resource_tests
from keystone.tests.unit.trust import test_backends as trust_tests
from keystone.trust.backends import sql as trust_sql

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class SqlTests(unit.SQLDriverOverrides, unit.TestCase):
    def setUp(self):
        super().setUp()
        self.database_fixture = self.useFixture(database.Database())
        self.load_backends()

        # populate the engine with tables & fixtures
        self.load_fixtures(default_fixtures)
        # defaulted by the data load
        self.user_foo['enabled'] = True

    def config_files(self):
        config_files = super().config_files()
        config_files.append(unit.dirs.tests_conf('backend_sql.conf'))
        return config_files


class DataTypeRoundTrips(SqlTests):
    def test_json_blob_roundtrip(self):
        """Test round-trip of a JSON data structure with JsonBlob."""
        with sql.session_for_read() as session:
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.literal({"key": "value"}, type_=core.JsonBlob)
                )
            )

        self.assertEqual({"key": "value"}, val)

    def test_json_blob_sql_null(self):
        """Test that JsonBlob can accommodate a SQL NULL value in a result set.

        SQL NULL may be handled by JsonBlob in the case where a table is
        storing NULL in a JsonBlob column, as several models use this type
        in a column that is nullable.   It also comes back when the column
        is left NULL from being in an OUTER JOIN.  In Python, this means
        the None constant is handled by the datatype.

        """
        with sql.session_for_read() as session:
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.cast(sqlalchemy.null(), type_=core.JsonBlob)
                )
            )

        self.assertIsNone(val)

    def test_json_blob_python_none(self):
        """Test that JsonBlob round-trips a Python None.

        This is where JSON datatypes get a little nutty, in that JSON has
        a 'null' keyword, and JsonBlob right now will persist Python None
        as the json string 'null', not SQL NULL.

        """
        with sql.session_for_read() as session:
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.literal(None, type_=core.JsonBlob)
                )
            )

        self.assertIsNone(val)

    def test_json_blob_python_none_renders(self):
        """Test that JsonBlob actually renders JSON 'null' for Python None."""
        with sql.session_for_read() as session:
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.cast(
                        sqlalchemy.literal(None, type_=core.JsonBlob),
                        sqlalchemy.String,
                    )
                )
            )

        self.assertEqual("null", val)

    def test_datetimeint_roundtrip(self):
        """Test round-trip of a Python datetime with DateTimeInt."""
        with sql.session_for_read() as session:
            datetime_value = datetime.datetime(2019, 5, 15, 10, 17, 55)
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.literal(datetime_value, type_=core.DateTimeInt)
                )
            )

        self.assertEqual(datetime_value, val)

    def test_datetimeint_persistence(self):
        """Test integer persistence with DateTimeInt."""
        with sql.session_for_read() as session:
            datetime_value = datetime.datetime(2019, 5, 15, 10, 17, 55)
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.cast(
                        sqlalchemy.literal(
                            datetime_value, type_=core.DateTimeInt
                        ),
                        sqlalchemy.Integer,
                    )
                )
            )

        self.assertEqual(1557915475000000, val)

    def test_datetimeint_python_none(self):
        """Test round-trip of a Python None with DateTimeInt."""
        with sql.session_for_read() as session:
            val = session.scalar(
                sqlalchemy.select(
                    sqlalchemy.literal(None, type_=core.DateTimeInt)
                )
            )

        self.assertIsNone(val)


class SqlModels(SqlTests):
    def load_table(self, name):
        table = sqlalchemy.Table(
            name,
            sql.ModelBase.metadata,
            autoload_with=self.database_fixture.engine,
        )
        return table

    def assertExpectedSchema(self, table, expected_schema):
        """Assert that a table's schema is what we expect.

        :param string table: the name of the table to inspect
        :param tuple expected_schema: a tuple of tuples containing the
            expected schema
        :raises AssertionError: when the database schema doesn't match the
            expected schema

        The expected_schema format is simply::

            (
                ('column name', sql type, qualifying detail),
                ...
            )

        The qualifying detail varies based on the type of the column::

          - sql.Boolean columns must indicate the column's default value or
            None if there is no default
          - Columns with a length, like sql.String, must indicate the
            column's length
          - All other column types should use None

        Example::

            cols = (
                ('id', sql.String, 64),
                ('enabled', sql.Boolean, True),
                ('extra', sql.JsonBlob, None),
            )
            self.assertExpectedSchema('table_name', cols)

        """
        table = self.load_table(table)

        actual_schema = []
        for column in table.c:
            if isinstance(column.type, sql.Boolean):
                default = None
                if column.default:
                    default = column.default.arg
                actual_schema.append((column.name, type(column.type), default))
            elif hasattr(column.type, 'length') and not isinstance(
                column.type, sql.Enum
            ):
                # NOTE(dstanek): Even though sql.Enum columns have a length
                # set we don't want to catch them here. Maybe in the future
                # we'll check to see that they contain a list of the correct
                # possible values.
                actual_schema.append(
                    (column.name, type(column.type), column.type.length)
                )
            else:
                actual_schema.append((column.name, type(column.type), None))

        self.assertCountEqual(expected_schema, actual_schema)

    def test_user_model(self):
        cols = (
            ('id', sql.String, 64),
            ('domain_id', sql.String, 64),
            ('default_project_id', sql.String, 64),
            ('enabled', sql.Boolean, None),
            ('extra', sql.JsonBlob, None),
            ('created_at', sql.DateTime, None),
            ('last_active_at', sqlalchemy.Date, None),
        )
        self.assertExpectedSchema('user', cols)

    def test_local_user_model(self):
        cols = (
            ('id', sql.Integer, None),
            ('user_id', sql.String, 64),
            ('name', sql.String, 255),
            ('domain_id', sql.String, 64),
            ('failed_auth_count', sql.Integer, None),
            ('failed_auth_at', sql.DateTime, None),
        )
        self.assertExpectedSchema('local_user', cols)

    def test_password_model(self):
        cols = (
            ('id', sql.Integer, None),
            ('local_user_id', sql.Integer, None),
            ('password_hash', sql.String, 255),
            ('created_at', sql.DateTime, None),
            ('expires_at', sql.DateTime, None),
            ('created_at_int', sql.DateTimeInt, None),
            ('expires_at_int', sql.DateTimeInt, None),
            ('self_service', sql.Boolean, False),
        )
        self.assertExpectedSchema('password', cols)

    def test_federated_user_model(self):
        cols = (
            ('id', sql.Integer, None),
            ('user_id', sql.String, 64),
            ('idp_id', sql.String, 64),
            ('protocol_id', sql.String, 64),
            ('unique_id', sql.String, 255),
            ('display_name', sql.String, 255),
        )
        self.assertExpectedSchema('federated_user', cols)

    def test_nonlocal_user_model(self):
        cols = (
            ('domain_id', sql.String, 64),
            ('name', sql.String, 255),
            ('user_id', sql.String, 64),
        )
        self.assertExpectedSchema('nonlocal_user', cols)

    def test_group_model(self):
        cols = (
            ('id', sql.String, 64),
            ('name', sql.String, 64),
            ('description', sql.Text, None),
            ('domain_id', sql.String, 64),
            ('extra', sql.JsonBlob, None),
        )
        self.assertExpectedSchema('group', cols)

    def test_project_model(self):
        cols = (
            ('id', sql.String, 64),
            ('name', sql.String, 64),
            ('description', sql.Text, None),
            ('domain_id', sql.String, 64),
            ('enabled', sql.Boolean, None),
            ('extra', sql.JsonBlob, None),
            ('parent_id', sql.String, 64),
            ('is_domain', sql.Boolean, False),
        )
        self.assertExpectedSchema('project', cols)

    def test_role_assignment_model(self):
        cols = (
            ('type', sql.Enum, None),
            ('actor_id', sql.String, 64),
            ('target_id', sql.String, 64),
            ('role_id', sql.String, 64),
            ('inherited', sql.Boolean, False),
        )
        self.assertExpectedSchema('assignment', cols)

    def test_user_group_membership(self):
        cols = (('group_id', sql.String, 64), ('user_id', sql.String, 64))
        self.assertExpectedSchema('user_group_membership', cols)

    def test_revocation_event_model(self):
        cols = (
            ('id', sql.Integer, None),
            ('domain_id', sql.String, 64),
            ('project_id', sql.String, 64),
            ('user_id', sql.String, 64),
            ('role_id', sql.String, 64),
            ('trust_id', sql.String, 64),
            ('consumer_id', sql.String, 64),
            ('access_token_id', sql.String, 64),
            ('issued_before', sql.DateTime, None),
            ('expires_at', sql.DateTime, None),
            ('revoked_at', sql.DateTime, None),
            ('audit_id', sql.String, 32),
            ('audit_chain_id', sql.String, 32),
        )
        self.assertExpectedSchema('revocation_event', cols)

    def test_project_tags_model(self):
        cols = (('project_id', sql.String, 64), ('name', sql.Unicode, 255))
        self.assertExpectedSchema('project_tag', cols)


class SqlIdentity(
    SqlTests,
    identity_tests.IdentityTests,
    assignment_tests.AssignmentTests,
    assignment_tests.SystemAssignmentTests,
    resource_tests.ResourceTests,
):
    def test_password_hashed(self):
        with sql.session_for_read() as session:
            user_ref = PROVIDERS.identity_api._get_user(
                session, self.user_foo['id']
            )
            self.assertNotEqual(
                self.user_foo['password'], user_ref['password']
            )

    def test_create_user_with_null_password(self):
        user_dict = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        user_dict["password"] = None
        new_user_dict = PROVIDERS.identity_api.create_user(user_dict)
        with sql.session_for_read() as session:
            new_user_ref = PROVIDERS.identity_api._get_user(
                session, new_user_dict['id']
            )
            self.assertIsNone(new_user_ref.password)

    def test_update_user_with_null_password(self):
        user_dict = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.assertTrue(user_dict['password'])
        new_user_dict = PROVIDERS.identity_api.create_user(user_dict)
        new_user_dict["password"] = None
        new_user_dict = PROVIDERS.identity_api.update_user(
            new_user_dict['id'], new_user_dict
        )
        with sql.session_for_read() as session:
            new_user_ref = PROVIDERS.identity_api._get_user(
                session, new_user_dict['id']
            )
            self.assertIsNone(new_user_ref.password)

    def test_delete_user_with_project_association(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        role_member = unit.new_role_ref()
        PROVIDERS.role_api.create_role(role_member['id'], role_member)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user['id'], self.project_bar['id'], role_member['id']
        )
        PROVIDERS.identity_api.delete_user(user['id'])
        self.assertRaises(
            exception.UserNotFound,
            PROVIDERS.assignment_api.list_projects_for_user,
            user['id'],
        )

    def test_create_user_case_sensitivity(self):
        # user name case sensitivity is down to the fact that it is marked as
        # an SQL UNIQUE column, which may not be valid for other backends, like
        # LDAP.

        # create a ref with a lowercase name
        ref = unit.new_user_ref(
            name=uuid.uuid4().hex.lower(),
            domain_id=CONF.identity.default_domain_id,
        )
        ref = PROVIDERS.identity_api.create_user(ref)

        # assign a new ID with the same name, but this time in uppercase
        ref['name'] = ref['name'].upper()
        PROVIDERS.identity_api.create_user(ref)

    def test_create_project_case_sensitivity(self):
        # project name case sensitivity is down to the fact that it is marked
        # as an SQL UNIQUE column, which may not be valid for other backends,
        # like LDAP.

        # create a ref with a lowercase name
        ref = unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        PROVIDERS.resource_api.create_project(ref['id'], ref)

        # assign a new ID with the same name, but this time in uppercase
        ref['id'] = uuid.uuid4().hex
        ref['name'] = ref['name'].upper()
        PROVIDERS.resource_api.create_project(ref['id'], ref)

    def test_delete_project_with_user_association(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        role_member = unit.new_role_ref()
        PROVIDERS.role_api.create_role(role_member['id'], role_member)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user['id'], self.project_bar['id'], role_member['id']
        )
        PROVIDERS.resource_api.delete_project(self.project_bar['id'])
        projects = PROVIDERS.assignment_api.list_projects_for_user(user['id'])
        self.assertEqual([], projects)

    def test_update_project_returns_extra(self):
        """Test for backward compatibility with an essex/folsom bug.

        Non-indexed attributes were returned in an 'extra' attribute, instead
        of on the entity itself; for consistency and backwards compatibility,
        those attributes should be included twice.

        This behavior is specific to the SQL driver.

        """
        arbitrary_key = uuid.uuid4().hex
        arbitrary_value = uuid.uuid4().hex
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project[arbitrary_key] = arbitrary_value
        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertEqual(arbitrary_value, ref[arbitrary_key])
        self.assertNotIn('extra', ref)

        ref['name'] = uuid.uuid4().hex
        ref = PROVIDERS.resource_api.update_project(ref['id'], ref)
        self.assertEqual(arbitrary_value, ref[arbitrary_key])
        self.assertEqual(arbitrary_value, ref['extra'][arbitrary_key])

    def test_update_user_returns_extra(self):
        """Test for backwards-compatibility with an essex/folsom bug.

        Non-indexed attributes were returned in an 'extra' attribute, instead
        of on the entity itself; for consistency and backwards compatibility,
        those attributes should be included twice.

        This behavior is specific to the SQL driver.

        """
        arbitrary_key = uuid.uuid4().hex
        arbitrary_value = uuid.uuid4().hex
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user[arbitrary_key] = arbitrary_value
        del user["id"]
        ref = PROVIDERS.identity_api.create_user(user)
        self.assertEqual(arbitrary_value, ref[arbitrary_key])
        self.assertNotIn('password', ref)
        self.assertNotIn('extra', ref)

        user['name'] = uuid.uuid4().hex
        user['password'] = uuid.uuid4().hex
        ref = PROVIDERS.identity_api.update_user(ref['id'], user)
        self.assertNotIn('password', ref)
        self.assertNotIn('password', ref['extra'])
        self.assertEqual(arbitrary_value, ref[arbitrary_key])
        self.assertEqual(arbitrary_value, ref['extra'][arbitrary_key])

    def test_sql_user_to_dict_null_default_project_id(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with sql.session_for_read() as session:
            query = session.query(identity_sql.User)
            query = query.filter_by(id=user['id'])
            raw_user_ref = query.one()
            self.assertIsNone(raw_user_ref.default_project_id)
            user_ref = raw_user_ref.to_dict()
            self.assertNotIn('default_project_id', user_ref)
            session.close()

    def test_list_domains_for_user(self):
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        user = unit.new_user_ref(domain_id=domain['id'])

        test_domain1 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(test_domain1['id'], test_domain1)
        test_domain2 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(test_domain2['id'], test_domain2)

        user = PROVIDERS.identity_api.create_user(user)
        user_domains = PROVIDERS.assignment_api.list_domains_for_user(
            user['id']
        )
        self.assertEqual(0, len(user_domains))
        PROVIDERS.assignment_api.create_grant(
            user_id=user['id'],
            domain_id=test_domain1['id'],
            role_id=self.role_member['id'],
        )
        PROVIDERS.assignment_api.create_grant(
            user_id=user['id'],
            domain_id=test_domain2['id'],
            role_id=self.role_member['id'],
        )
        user_domains = PROVIDERS.assignment_api.list_domains_for_user(
            user['id']
        )
        self.assertThat(user_domains, matchers.HasLength(2))

    def test_list_domains_for_user_with_grants(self):
        # Create two groups each with a role on a different domain, and
        # make user1 a member of both groups.  Both these new domains
        # should now be included, along with any direct user grants.
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        user = unit.new_user_ref(domain_id=domain['id'])
        user = PROVIDERS.identity_api.create_user(user)
        group1 = unit.new_group_ref(domain_id=domain['id'])
        group1 = PROVIDERS.identity_api.create_group(group1)
        group2 = unit.new_group_ref(domain_id=domain['id'])
        group2 = PROVIDERS.identity_api.create_group(group2)

        test_domain1 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(test_domain1['id'], test_domain1)
        test_domain2 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(test_domain2['id'], test_domain2)
        test_domain3 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(test_domain3['id'], test_domain3)

        PROVIDERS.identity_api.add_user_to_group(user['id'], group1['id'])
        PROVIDERS.identity_api.add_user_to_group(user['id'], group2['id'])

        # Create 3 grants, one user grant, the other two as group grants
        PROVIDERS.assignment_api.create_grant(
            user_id=user['id'],
            domain_id=test_domain1['id'],
            role_id=self.role_member['id'],
        )
        PROVIDERS.assignment_api.create_grant(
            group_id=group1['id'],
            domain_id=test_domain2['id'],
            role_id=self.role_admin['id'],
        )
        PROVIDERS.assignment_api.create_grant(
            group_id=group2['id'],
            domain_id=test_domain3['id'],
            role_id=self.role_admin['id'],
        )
        user_domains = PROVIDERS.assignment_api.list_domains_for_user(
            user['id']
        )
        self.assertThat(user_domains, matchers.HasLength(3))

    def test_list_domains_for_user_with_inherited_grants(self):
        """Test that inherited roles on the domain are excluded.

        Test Plan:

        - Create two domains, one user, group and role
        - Domain1 is given an inherited user role, Domain2 an inherited
          group role (for a group of which the user is a member)
        - When listing domains for user, neither domain should be returned

        """
        domain1 = unit.new_domain_ref()
        domain1 = PROVIDERS.resource_api.create_domain(domain1['id'], domain1)
        domain2 = unit.new_domain_ref()
        domain2 = PROVIDERS.resource_api.create_domain(domain2['id'], domain2)
        user = unit.new_user_ref(domain_id=domain1['id'])
        user = PROVIDERS.identity_api.create_user(user)
        group = unit.new_group_ref(domain_id=domain1['id'])
        group = PROVIDERS.identity_api.create_group(group)
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        role = unit.new_role_ref()
        PROVIDERS.role_api.create_role(role['id'], role)

        # Create a grant on each domain, one user grant, one group grant,
        # both inherited.
        PROVIDERS.assignment_api.create_grant(
            user_id=user['id'],
            domain_id=domain1['id'],
            role_id=role['id'],
            inherited_to_projects=True,
        )
        PROVIDERS.assignment_api.create_grant(
            group_id=group['id'],
            domain_id=domain2['id'],
            role_id=role['id'],
            inherited_to_projects=True,
        )

        user_domains = PROVIDERS.assignment_api.list_domains_for_user(
            user['id']
        )
        # No domains should be returned since both domains have only inherited
        # roles assignments.
        self.assertThat(user_domains, matchers.HasLength(0))

    def test_list_groups_for_user(self):
        domain = self._get_domain_fixture()
        test_groups = []
        test_users = []
        GROUP_COUNT = 3
        USER_COUNT = 2

        for x in range(0, USER_COUNT):
            new_user = unit.new_user_ref(domain_id=domain['id'])
            new_user = PROVIDERS.identity_api.create_user(new_user)
            test_users.append(new_user)
        positive_user = test_users[0]
        negative_user = test_users[1]

        for x in range(0, USER_COUNT):
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                test_users[x]['id']
            )
            self.assertEqual(0, len(group_refs))

        for x in range(0, GROUP_COUNT):
            before_count = x
            after_count = x + 1
            new_group = unit.new_group_ref(domain_id=domain['id'])
            new_group = PROVIDERS.identity_api.create_group(new_group)
            test_groups.append(new_group)

            # add the user to the group and ensure that the
            # group count increases by one for each
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id']
            )
            self.assertEqual(before_count, len(group_refs))
            PROVIDERS.identity_api.add_user_to_group(
                positive_user['id'], new_group['id']
            )
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id']
            )
            self.assertEqual(after_count, len(group_refs))

            # Make sure the group count for the unrelated user did not change
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                negative_user['id']
            )
            self.assertEqual(0, len(group_refs))

        # remove the user from each group and ensure that
        # the group count reduces by one for each
        for x in range(0, 3):
            before_count = GROUP_COUNT - x
            after_count = GROUP_COUNT - x - 1
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id']
            )
            self.assertEqual(before_count, len(group_refs))
            PROVIDERS.identity_api.remove_user_from_group(
                positive_user['id'], test_groups[x]['id']
            )
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                positive_user['id']
            )
            self.assertEqual(after_count, len(group_refs))
            # Make sure the group count for the unrelated user
            # did not change
            group_refs = PROVIDERS.identity_api.list_groups_for_user(
                negative_user['id']
            )
            self.assertEqual(0, len(group_refs))

    def test_add_user_to_group_expiring_mapped(self):
        self._build_fed_resource()
        domain = self._get_domain_fixture()
        self.config_fixture.config(
            group='federation', default_authorization_ttl=5
        )
        time = timeutils.utcnow()
        tick = datetime.timedelta(minutes=5)

        new_group = unit.new_group_ref(domain_id=domain['id'])
        new_group = PROVIDERS.identity_api.create_group(new_group)

        fed_dict = unit.new_federated_user_ref()
        fed_dict['id'] = fed_dict['unique_id']
        fed_dict['name'] = fed_dict['display_name']
        fed_dict['domain'] = {'id': uuid.uuid4().hex}

        fed_dict['idp_id'] = 'myidp'
        fed_dict['protocol_id'] = 'mapped'

        with freezegun.freeze_time(time - tick) as frozen_time:
            user = PROVIDERS.identity_api.shadow_federated_user(
                fed_dict['idp_id'],
                fed_dict['protocol_id'],
                fed_dict,
                group_ids=[new_group['id']],
            )

            PROVIDERS.identity_api.check_user_in_group(
                user['id'], new_group['id']
            )

            # Expiration
            frozen_time.tick(tick)
            self.assertRaises(
                exception.NotFound,
                PROVIDERS.identity_api.check_user_in_group,
                user['id'],
                new_group['id'],
            )

            # Renewal
            PROVIDERS.identity_api.shadow_federated_user(
                fed_dict['idp_id'],
                fed_dict['protocol_id'],
                fed_dict,
                group_ids=[new_group['id']],
            )
            PROVIDERS.identity_api.check_user_in_group(
                user['id'], new_group['id']
            )

    def test_add_user_to_group_expiring(self):
        self._build_fed_resource()
        domain = self._get_domain_fixture()
        time = timeutils.utcnow()
        tick = datetime.timedelta(minutes=5)

        new_group = unit.new_group_ref(domain_id=domain['id'])
        new_group = PROVIDERS.identity_api.create_group(new_group)

        fed_dict = unit.new_federated_user_ref()
        fed_dict['idp_id'] = 'myidp'
        fed_dict['protocol_id'] = 'mapped'
        new_user = PROVIDERS.shadow_users_api.create_federated_user(
            domain['id'], fed_dict
        )

        with freezegun.freeze_time(time - tick) as frozen_time:
            PROVIDERS.shadow_users_api.add_user_to_group_expires(
                new_user['id'], new_group['id']
            )

            self.config_fixture.config(
                group='federation', default_authorization_ttl=0
            )
            self.assertRaises(
                exception.NotFound,
                PROVIDERS.identity_api.check_user_in_group,
                new_user['id'],
                new_group['id'],
            )

            self.config_fixture.config(
                group='federation', default_authorization_ttl=5
            )
            PROVIDERS.identity_api.check_user_in_group(
                new_user['id'], new_group['id']
            )

            # Expiration
            frozen_time.tick(tick)
            self.assertRaises(
                exception.NotFound,
                PROVIDERS.identity_api.check_user_in_group,
                new_user['id'],
                new_group['id'],
            )

            # Renewal
            PROVIDERS.shadow_users_api.add_user_to_group_expires(
                new_user['id'], new_group['id']
            )
            PROVIDERS.identity_api.check_user_in_group(
                new_user['id'], new_group['id']
            )

    def test_add_user_to_group_expiring_list(self):
        self._build_fed_resource()
        domain = self._get_domain_fixture()
        self.config_fixture.config(
            group='federation', default_authorization_ttl=5
        )
        time = timeutils.utcnow()
        tick = datetime.timedelta(minutes=5)

        new_group = unit.new_group_ref(domain_id=domain['id'])
        new_group = PROVIDERS.identity_api.create_group(new_group)
        exp_new_group = unit.new_group_ref(domain_id=domain['id'])
        exp_new_group = PROVIDERS.identity_api.create_group(exp_new_group)

        fed_dict = unit.new_federated_user_ref()
        fed_dict['idp_id'] = 'myidp'
        fed_dict['protocol_id'] = 'mapped'
        new_user = PROVIDERS.shadow_users_api.create_federated_user(
            domain['id'], fed_dict
        )

        PROVIDERS.identity_api.add_user_to_group(
            new_user['id'], new_group['id']
        )
        PROVIDERS.identity_api.check_user_in_group(
            new_user['id'], new_group['id']
        )

        with freezegun.freeze_time(time - tick) as frozen_time:
            PROVIDERS.shadow_users_api.add_user_to_group_expires(
                new_user['id'], exp_new_group['id']
            )
            PROVIDERS.identity_api.check_user_in_group(
                new_user['id'], new_group['id']
            )

            groups = PROVIDERS.identity_api.list_groups_for_user(
                new_user['id']
            )
            self.assertEqual(len(groups), 2)
            for group in groups:
                if group.get('membership_expires_at'):
                    self.assertEqual(group['membership_expires_at'], time)

            frozen_time.tick(tick)
            groups = PROVIDERS.identity_api.list_groups_for_user(
                new_user['id']
            )
            self.assertEqual(len(groups), 1)

    def test_storing_null_domain_id_in_project_ref(self):
        """Test the special storage of domain_id=None in sql resource driver.

        The resource driver uses a special value in place of None for domain_id
        in the project record. This shouldn't escape the driver. Hence we test
        the interface to ensure that you can store a domain_id of None, and
        that any special value used inside the driver does not escape through
        the interface.

        """
        spoiler_project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(
            spoiler_project['id'], spoiler_project
        )

        # First let's create a project with a None domain_id and make sure we
        # can read it back.
        project = unit.new_project_ref(domain_id=None, is_domain=True)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertDictEqual(project, ref)

        # Can we get it by name?
        ref = PROVIDERS.resource_api.get_project_by_name(project['name'], None)
        self.assertDictEqual(project, ref)

        # Can we filter for them - create a second domain to ensure we are
        # testing the receipt of more than one.
        project2 = unit.new_project_ref(domain_id=None, is_domain=True)
        project2 = PROVIDERS.resource_api.create_project(
            project2['id'], project2
        )
        hints = driver_hints.Hints()
        hints.add_filter('domain_id', None)
        refs = PROVIDERS.resource_api.list_projects(hints)
        self.assertThat(refs, matchers.HasLength(2 + self.domain_count))
        self.assertIn(project, refs)
        self.assertIn(project2, refs)

        # Can we update it?
        project['name'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_project(project['id'], project)
        ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertDictEqual(project, ref)

        # Finally, make sure we can delete it
        project['enabled'] = False
        PROVIDERS.resource_api.update_project(project['id'], project)
        PROVIDERS.resource_api.delete_project(project['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            project['id'],
        )

    def test_hidden_project_domain_root_is_really_hidden(self):
        """Ensure we cannot access the hidden root of all project domains.

        Calling any of the driver methods should result in the same as
        would be returned if we passed a project that does not exist. We don't
        test create_project, since we do not allow a caller of our API to
        specify their own ID for a new entity.

        """

        def _exercise_project_api(ref_id):
            driver = PROVIDERS.resource_api.driver
            self.assertRaises(
                exception.ProjectNotFound, driver.get_project, ref_id
            )

            self.assertRaises(
                exception.ProjectNotFound,
                driver.get_project_by_name,
                resource.NULL_DOMAIN_ID,
                ref_id,
            )

            project_ids = [
                x['id'] for x in driver.list_projects(driver_hints.Hints())
            ]
            self.assertNotIn(ref_id, project_ids)

            projects = driver.list_projects_from_ids([ref_id])
            self.assertThat(projects, matchers.HasLength(0))

            project_ids = list(
                driver.list_project_ids_from_domain_ids([ref_id])
            )
            self.assertNotIn(ref_id, project_ids)

            self.assertRaises(
                exception.DomainNotFound,
                driver.list_projects_in_domain,
                ref_id,
            )

            project_ids = [
                x['id']
                for x in driver.list_projects_acting_as_domain(
                    driver_hints.Hints()
                )
            ]
            self.assertNotIn(ref_id, project_ids)

            projects = driver.list_projects_in_subtree(ref_id)
            self.assertThat(projects, matchers.HasLength(0))

            self.assertRaises(
                exception.ProjectNotFound, driver.list_project_parents, ref_id
            )

            # A non-existing project just returns True from the driver
            self.assertTrue(driver.is_leaf_project(ref_id))

            self.assertRaises(
                exception.ProjectNotFound, driver.update_project, ref_id, {}
            )

            self.assertRaises(
                exception.ProjectNotFound, driver.delete_project, ref_id
            )

            # Deleting list of projects that includes a non-existing project
            # should be silent. The root domain <<keystone.domain.root>> can't
            # be deleted.
            if ref_id != resource.NULL_DOMAIN_ID:
                driver.delete_projects_from_ids([ref_id])

        _exercise_project_api(uuid.uuid4().hex)
        _exercise_project_api(resource.NULL_DOMAIN_ID)

    def test_list_users_call_count(self):
        """There should not be O(N) queries."""
        # create 10 users. 10 is just a random number
        for i in range(10):
            user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
            PROVIDERS.identity_api.create_user(user)

        # sqlalchemy emits various events and allows to listen to them. Here
        # bound method `query_counter` will be called each time when a query
        # is compiled
        class CallCounter:
            def __init__(self):
                self.calls = 0

            def reset(self):
                self.calls = 0

            def query_counter(self, query):
                self.calls += 1

        counter = CallCounter()
        sqlalchemy.event.listen(
            sqlalchemy.orm.query.Query, 'before_compile', counter.query_counter
        )

        first_call_users = PROVIDERS.identity_api.list_users()
        first_call_counter = counter.calls
        # add 10 more users
        for i in range(10):
            user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
            PROVIDERS.identity_api.create_user(user)
        counter.reset()
        second_call_users = PROVIDERS.identity_api.list_users()
        # ensure that the number of calls does not depend on the number of
        # users fetched.
        self.assertNotEqual(len(first_call_users), len(second_call_users))
        self.assertEqual(first_call_counter, counter.calls)
        self.assertEqual(3, counter.calls)

    def test_check_project_depth(self):
        # Create a 3 level project tree:
        #
        # default_domain
        #       |
        #   project_1
        #       |
        #   project_2
        project_1 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project_1['id'], project_1)
        project_2 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id,
            parent_id=project_1['id'],
        )
        PROVIDERS.resource_api.create_project(project_2['id'], project_2)

        # if max_depth is None or >= current project depth, return nothing.
        resp = PROVIDERS.resource_api.check_project_depth(max_depth=None)
        self.assertIsNone(resp)
        resp = PROVIDERS.resource_api.check_project_depth(max_depth=3)
        self.assertIsNone(resp)
        resp = PROVIDERS.resource_api.check_project_depth(max_depth=4)
        self.assertIsNone(resp)
        # if max_depth < current project depth, raise LimitTreeExceedError
        self.assertRaises(
            exception.LimitTreeExceedError,
            PROVIDERS.resource_api.check_project_depth,
            2,
        )

    def test_update_user_with_stale_data_forces_retry(self):
        # Capture log output so we know oslo.db attempted a retry
        log_fixture = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))

        # Create a new user
        user_dict = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        new_user_dict = PROVIDERS.identity_api.create_user(user_dict)

        side_effects = [
            # Raise a StaleDataError simulating that another client has
            # updated the user's password while this client's request was
            # being processed
            sqlalchemy.orm.exc.StaleDataError,
            # The oslo.db library will retry the request, so the second
            # time this method is called let's return a valid session
            # object
            sql.session_for_write(),
        ]
        with mock.patch('keystone.common.sql.session_for_write') as m:
            m.side_effect = side_effects

            # Update a user's attribute, the first attempt will fail but
            # oslo.db will handle the exception and retry, the second attempt
            # will succeed
            new_user_dict['email'] = uuid.uuid4().hex
            PROVIDERS.identity_api.update_user(
                new_user_dict['id'], new_user_dict
            )

        # Make sure oslo.db retried the update by checking the log output
        expected_log_message = (
            'Performing DB retry for function keystone.identity.backends.'
            'sql.Identity.update_user'
        )
        self.assertIn(expected_log_message, log_fixture.output)


class SqlTrust(SqlTests, trust_tests.TrustTests):
    def test_trust_expires_at_int_matches_expires_at(self):
        with sql.session_for_write() as session:
            new_id = uuid.uuid4().hex
            self.create_sample_trust(new_id)
            trust_ref = session.get(trust_sql.TrustModel, new_id)
            self.assertIsNotNone(trust_ref._expires_at)
            self.assertEqual(trust_ref._expires_at, trust_ref.expires_at_int)
            self.assertEqual(trust_ref.expires_at, trust_ref.expires_at_int)


class SqlCatalog(SqlTests, catalog_tests.CatalogTests):
    _legacy_endpoint_id_in_endpoint = True
    _enabled_default_to_true_when_creating_endpoint = True

    def test_get_v3_catalog_project_non_exist(self):
        service = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(service['id'], service)

        malformed_url = "http://192.168.1.104:8774/v2/$(project)s"
        endpoint = unit.new_endpoint_ref(
            service_id=service['id'], url=malformed_url, region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint.copy())
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.catalog_api.get_v3_catalog,
            'fake-user',
            'fake-project',
        )

    def test_get_v3_catalog_with_empty_public_url(self):
        service = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(service['id'], service)

        endpoint = unit.new_endpoint_ref(
            url='', service_id=service['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint.copy())

        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            self.user_foo['id'], self.project_bar['id']
        )
        catalog_endpoint = catalog[0]
        self.assertEqual(service['name'], catalog_endpoint['name'])
        self.assertEqual(service['id'], catalog_endpoint['id'])
        self.assertEqual([], catalog_endpoint['endpoints'])

    def test_create_endpoint_region_returns_not_found(self):
        service = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(service['id'], service)

        endpoint = unit.new_endpoint_ref(
            region_id=uuid.uuid4().hex, service_id=service['id']
        )

        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.catalog_api.create_endpoint,
            endpoint['id'],
            endpoint.copy(),
        )

    def test_create_region_invalid_id(self):
        region = unit.new_region_ref(id='0' * 256)

        self.assertRaises(
            exception.StringLengthExceeded,
            PROVIDERS.catalog_api.create_region,
            region,
        )

    def test_create_region_invalid_parent_id(self):
        region = unit.new_region_ref(parent_region_id='0' * 256)

        self.assertRaises(
            exception.RegionNotFound,
            PROVIDERS.catalog_api.create_region,
            region,
        )

    def test_delete_region_with_endpoint(self):
        # create a region
        region = unit.new_region_ref()
        PROVIDERS.catalog_api.create_region(region)

        # create a child region
        child_region = unit.new_region_ref(parent_region_id=region['id'])
        PROVIDERS.catalog_api.create_region(child_region)
        # create a service
        service = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(service['id'], service)

        # create an endpoint attached to the service and child region
        child_endpoint = unit.new_endpoint_ref(
            region_id=child_region['id'], service_id=service['id']
        )

        PROVIDERS.catalog_api.create_endpoint(
            child_endpoint['id'], child_endpoint
        )
        self.assertRaises(
            exception.RegionDeletionError,
            PROVIDERS.catalog_api.delete_region,
            child_region['id'],
        )

        # create an endpoint attached to the service and parent region
        endpoint = unit.new_endpoint_ref(
            region_id=region['id'], service_id=service['id']
        )

        PROVIDERS.catalog_api.create_endpoint(endpoint['id'], endpoint)
        self.assertRaises(
            exception.RegionDeletionError,
            PROVIDERS.catalog_api.delete_region,
            region['id'],
        )

    def test_v3_catalog_domain_scoped_token(self):
        # test the case that project_id is None.
        srv_1 = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(srv_1['id'], srv_1)
        endpoint_1 = unit.new_endpoint_ref(
            service_id=srv_1['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_1['id'], endpoint_1)

        srv_2 = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(srv_2['id'], srv_2)
        endpoint_2 = unit.new_endpoint_ref(
            service_id=srv_2['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_2['id'], endpoint_2)

        self.config_fixture.config(
            group='endpoint_filter', return_all_endpoints_if_no_filter=True
        )
        catalog_ref = PROVIDERS.catalog_api.get_v3_catalog(
            uuid.uuid4().hex, None
        )
        self.assertThat(catalog_ref, matchers.HasLength(2))
        self.config_fixture.config(
            group='endpoint_filter', return_all_endpoints_if_no_filter=False
        )
        catalog_ref = PROVIDERS.catalog_api.get_v3_catalog(
            uuid.uuid4().hex, None
        )
        self.assertThat(catalog_ref, matchers.HasLength(0))

    def test_v3_catalog_endpoint_filter_enabled(self):
        srv_1 = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(srv_1['id'], srv_1)
        endpoint_1 = unit.new_endpoint_ref(
            service_id=srv_1['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_1['id'], endpoint_1)
        endpoint_2 = unit.new_endpoint_ref(
            service_id=srv_1['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_2['id'], endpoint_2)
        # create endpoint-project association.
        PROVIDERS.catalog_api.add_endpoint_to_project(
            endpoint_1['id'], self.project_bar['id']
        )

        catalog_ref = PROVIDERS.catalog_api.get_v3_catalog(
            uuid.uuid4().hex, self.project_bar['id']
        )
        self.assertThat(catalog_ref, matchers.HasLength(1))
        self.assertThat(catalog_ref[0]['endpoints'], matchers.HasLength(1))
        # the endpoint is that defined in the endpoint-project association.
        self.assertEqual(
            endpoint_1['id'], catalog_ref[0]['endpoints'][0]['id']
        )

    def test_v3_catalog_endpoint_filter_disabled(self):
        # there is no endpoint-project association defined.
        self.config_fixture.config(
            group='endpoint_filter', return_all_endpoints_if_no_filter=True
        )
        srv_1 = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(srv_1['id'], srv_1)
        endpoint_1 = unit.new_endpoint_ref(
            service_id=srv_1['id'], region_id=None
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_1['id'], endpoint_1)

        srv_2 = unit.new_service_ref()
        PROVIDERS.catalog_api.create_service(srv_2['id'], srv_2)

        catalog_ref = PROVIDERS.catalog_api.get_v3_catalog(
            uuid.uuid4().hex, self.project_bar['id']
        )
        self.assertThat(catalog_ref, matchers.HasLength(2))
        srv_id_list = [catalog_ref[0]['id'], catalog_ref[1]['id']]
        self.assertCountEqual([srv_1['id'], srv_2['id']], srv_id_list)


class SqlPolicy(SqlTests, policy_tests.PolicyTests):
    pass


class SqlInheritance(SqlTests, assignment_tests.InheritanceTests):
    pass


class SqlImpliedRoles(SqlTests, assignment_tests.ImpliedRoleTests):
    pass


class SqlFilterTests(SqlTests, identity_tests.FilterTests):
    def clean_up_entities(self):
        """Clean up entity test data from Filter Test Cases."""
        for entity in ['user', 'group', 'project']:
            self._delete_test_data(entity, self.entity_list[entity])
            self._delete_test_data(entity, self.domain1_entity_list[entity])
        del self.entity_list
        del self.domain1_entity_list
        self.domain1['enabled'] = False
        PROVIDERS.resource_api.update_domain(self.domain1['id'], self.domain1)
        PROVIDERS.resource_api.delete_domain(self.domain1['id'])
        del self.domain1

    def test_list_entities_filtered_by_domain(self):
        # NOTE(henry-nash): This method is here rather than in
        # unit.identity.test_backends since any domain filtering with LDAP is
        # handled by the manager layer (and is already tested elsewhere) not at
        # the driver level.
        self.addCleanup(self.clean_up_entities)
        self.domain1 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(self.domain1['id'], self.domain1)

        self.entity_list = {}
        self.domain1_entity_list = {}
        for entity in ['user', 'group', 'project']:
            # Create 5 entities, 3 of which are in domain1
            DOMAIN1_ENTITIES = 3
            self.entity_list[entity] = self._create_test_data(entity, 2)
            self.domain1_entity_list[entity] = self._create_test_data(
                entity, DOMAIN1_ENTITIES, self.domain1['id']
            )

            # Should get back the DOMAIN1_ENTITIES in domain1
            hints = driver_hints.Hints()
            hints.add_filter('domain_id', self.domain1['id'])
            entities = self._list_entities(entity)(hints=hints)
            self.assertEqual(DOMAIN1_ENTITIES, len(entities))
            self._match_with_list(entities, self.domain1_entity_list[entity])
            # Check the driver has removed the filter from the list hints
            self.assertFalse(hints.get_exact_filter_by_name('domain_id'))

    def test_filter_sql_injection_attack(self):
        """Test against sql injection attack on filters.

        Test Plan:
        - Attempt to get all entities back by passing a two-term attribute
        - Attempt to piggyback filter to damage DB (e.g. drop table)

        """
        # Check we have some users
        users = PROVIDERS.identity_api.list_users()
        self.assertGreater(len(users), 0)

        hints = driver_hints.Hints()
        hints.add_filter('name', "anything' or 'x'='x")
        users = PROVIDERS.identity_api.list_users(hints=hints)
        self.assertEqual(0, len(users))

        # See if we can add a SQL command...use the group table instead of the
        # user table since 'user' is reserved word for SQLAlchemy.
        group = unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        group = PROVIDERS.identity_api.create_group(group)

        hints = driver_hints.Hints()
        hints.add_filter('name', "x'; drop table group")
        groups = PROVIDERS.identity_api.list_groups(hints=hints)
        self.assertEqual(0, len(groups))

        groups = PROVIDERS.identity_api.list_groups()
        self.assertGreater(len(groups), 0)


class SqlLimitTests(SqlTests, identity_tests.LimitTests):
    def setUp(self):
        super().setUp()
        identity_tests.LimitTests.setUp(self)


class FakeTable(sql.ModelBase):
    __tablename__ = 'test_table'
    col = sql.Column(sql.String(32), primary_key=True)

    @sql.handle_conflicts('keystone')
    def insert(self):
        raise db_exception.DBDuplicateEntry

    @sql.handle_conflicts('keystone')
    def update(self):
        raise db_exception.DBError(
            inner_exception=exc.IntegrityError('a', 'a', 'a')
        )

    @sql.handle_conflicts('keystone')
    def lookup(self):
        raise KeyError


class SqlDecorators(unit.TestCase):
    def test_initialization_fail(self):
        self.assertRaises(
            exception.StringLengthExceeded, FakeTable, col='a' * 64
        )

    def test_initialization(self):
        tt = FakeTable(col='a')
        self.assertEqual('a', tt.col)

    def test_conflict_happend(self):
        self.assertRaises(exception.Conflict, FakeTable().insert)
        self.assertRaises(exception.UnexpectedError, FakeTable().update)

    def test_not_conflict_error(self):
        self.assertRaises(KeyError, FakeTable().lookup)


class SqlModuleInitialization(unit.TestCase):
    @mock.patch.object(sql.core, 'CONF')
    @mock.patch.object(options, 'set_defaults')
    def test_initialize_module(self, set_defaults, CONF):
        sql.initialize()
        set_defaults.assert_called_with(
            CONF, connection='sqlite:///keystone.db'
        )


class SqlCredential(SqlTests):
    def _create_credential_with_user_id(self, user_id=uuid.uuid4().hex):
        credential = unit.new_credential_ref(
            user_id=user_id, extra=uuid.uuid4().hex, type=uuid.uuid4().hex
        )
        PROVIDERS.credential_api.create_credential(
            credential['id'], credential
        )
        return credential

    def _validateCredentialList(
        self, retrieved_credentials, expected_credentials
    ):
        self.assertEqual(len(expected_credentials), len(retrieved_credentials))
        retrived_ids = [c['id'] for c in retrieved_credentials]
        for cred in expected_credentials:
            self.assertIn(cred['id'], retrived_ids)

    def setUp(self):
        self.useFixture(database.Database())
        super().setUp()
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'credential',
                credential_provider.MAX_ACTIVE_KEYS,
            )
        )

        self.credentials = []
        for _ in range(3):
            self.credentials.append(self._create_credential_with_user_id())
        self.user_credentials = []
        for _ in range(3):
            cred = self._create_credential_with_user_id(self.user_foo['id'])
            self.user_credentials.append(cred)
            self.credentials.append(cred)

    def test_list_credentials(self):
        credentials = PROVIDERS.credential_api.list_credentials()
        self._validateCredentialList(credentials, self.credentials)
        # test filtering using hints
        hints = driver_hints.Hints()
        hints.add_filter('user_id', self.user_foo['id'])
        credentials = PROVIDERS.credential_api.list_credentials(hints)
        self._validateCredentialList(credentials, self.user_credentials)

    def test_list_credentials_for_user(self):
        credentials = PROVIDERS.credential_api.list_credentials_for_user(
            self.user_foo['id']
        )
        self._validateCredentialList(credentials, self.user_credentials)

    def test_list_credentials_for_user_and_type(self):
        cred = self.user_credentials[0]
        credentials = PROVIDERS.credential_api.list_credentials_for_user(
            self.user_foo['id'], type=cred['type']
        )
        self._validateCredentialList(credentials, [cred])

    def test_create_credential_is_encrypted_when_stored(self):
        credential = unit.new_credential_ref(user_id=uuid.uuid4().hex)
        credential_id = credential['id']
        returned_credential = PROVIDERS.credential_api.create_credential(
            credential_id, credential
        )

        # Make sure the `blob` is *not* encrypted when returned from the
        # credential API.
        self.assertEqual(returned_credential['blob'], credential['blob'])

        credential_from_backend = (
            PROVIDERS.credential_api.driver.get_credential(credential_id)
        )

        # Pull the credential directly from the backend, the `blob` should be
        # encrypted.
        self.assertNotEqual(
            credential_from_backend['encrypted_blob'], credential['blob']
        )

    def test_list_credentials_is_decrypted(self):
        credential = unit.new_credential_ref(user_id=uuid.uuid4().hex)
        credential_id = credential['id']

        created_credential = PROVIDERS.credential_api.create_credential(
            credential_id, credential
        )

        # Pull the credential directly from the backend, the `blob` should be
        # encrypted.
        credential_from_backend = (
            PROVIDERS.credential_api.driver.get_credential(credential_id)
        )
        self.assertNotEqual(
            credential_from_backend['encrypted_blob'], credential['blob']
        )

        # Make sure the `blob` values listed from the API are not encrypted.
        listed_credentials = PROVIDERS.credential_api.list_credentials()
        self.assertIn(created_credential, listed_credentials)


class SqlRegisteredLimit(SqlTests, limit_tests.RegisteredLimitTests):
    def setUp(self):
        super().setUp()

        fixtures_to_cleanup = []
        for service in default_fixtures.SERVICES:
            service_id = service['id']
            rv = PROVIDERS.catalog_api.create_service(service_id, service)
            attrname = service['extra']['name']
            setattr(self, attrname, rv)
            fixtures_to_cleanup.append(attrname)
        for region in default_fixtures.REGIONS:
            rv = PROVIDERS.catalog_api.create_region(region)
            attrname = region['id']
            setattr(self, attrname, rv)
            fixtures_to_cleanup.append(attrname)
        self.addCleanup(self.cleanup_instance(*fixtures_to_cleanup))


class SqlLimit(SqlTests, limit_tests.LimitTests):
    def setUp(self):
        super().setUp()

        fixtures_to_cleanup = []
        for service in default_fixtures.SERVICES:
            service_id = service['id']
            rv = PROVIDERS.catalog_api.create_service(service_id, service)
            attrname = service['extra']['name']
            setattr(self, attrname, rv)
            fixtures_to_cleanup.append(attrname)
        for region in default_fixtures.REGIONS:
            rv = PROVIDERS.catalog_api.create_region(region)
            attrname = region['id']
            setattr(self, attrname, rv)
            fixtures_to_cleanup.append(attrname)
        self.addCleanup(self.cleanup_instance(*fixtures_to_cleanup))

        registered_limit_1 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_one['id'],
            resource_name='volume',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_2 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='snapshot',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        registered_limit_3 = unit.new_registered_limit_ref(
            service_id=self.service_one['id'],
            region_id=self.region_two['id'],
            resource_name='backup',
            default_limit=10,
            id=uuid.uuid4().hex,
        )
        PROVIDERS.unified_limit_api.create_registered_limits(
            [registered_limit_1, registered_limit_2, registered_limit_3]
        )
